from datetime import datetime
from typing import Callable, Optional

from ..exceptions.device import WrongElementError
from .property import Property


class MultiLevelSwitchProperty(Property):
    """
    Object for multi level switches. It stores the multi level state and additional information that help displaying the state
    in the right context.

    :param element_uid: Element UID, something like devolo.Dimmer:hdm:ZWave:CBC56091/24#2
    :key value: Value the multi level switch has at time of creating this instance
    :type value: float
    :key switch_type: Type this switch is of, e.g. temperature
    :type switch_type: str
    :key max: Highest possible value, that can be set
    :type max: float
    :key min: Lowest possible value, that can be set
    :type min: float
    """

    def __init__(self, element_uid: str, setter: Callable, **kwargs):
        if not element_uid.startswith(("devolo.Blinds:",
                                       "devolo.Dimmer:",
                                       "devolo.MultiLevelSwitch:",
                                       "devolo.SirenMultiLevelSwitch:")):
            raise WrongElementError(f"{element_uid} is not a multi level switch.")

        super().__init__(element_uid=element_uid)
        self._setter = setter

        self._value: float = kwargs.pop("value", 0.0)
        self.switch_type: str = kwargs.pop("switch_type", "")
        self.max: float = kwargs.pop("max", 100.0)
        self.min: float = kwargs.pop("min", 0.0)

    @property
    def last_activity(self) -> datetime:
        """ Date and time the state of the multi level switch was last updated. """
        return super().last_activity

    @last_activity.setter
    def last_activity(self, timestamp: int):
        """ The gateway persists the last activity of some multi level switchs. They can be initialized with that value. """
        if timestamp != -1:
            self._last_activity = datetime.utcfromtimestamp(timestamp / 1000)
            self._logger.debug("last_activity of element_uid %s set to %s.", self.element_uid, self._last_activity)

    @property
    def unit(self) -> Optional[str]:
        """ Human readable unit of the property. Defaults to percent. """
        units = {
            "temperature": "°C",
            "tone": None,
        }
        return units.get(self.switch_type, "%")

    @property
    def value(self) -> float:
        """ Multi level value. """
        return self._value

    @value.setter
    def value(self, value: float):
        """ Update value of the multilevel value and set point in time of the last_activity. """
        self._value = value
        self._last_activity = datetime.now()
        self._logger.debug("Value of %s set to %s.", self.element_uid, value)

    def set(self, value: float):
        """
        Set the multilevel switch of the given element_uid to the given value.

        :param value: Value to set
        """
        if value > self.max or value < self.min:
            raise ValueError((f"Set value {value} is too {'low' if value < self.min else 'high'}. "
                              f"The min value is {self.min}. The max value is {self.max}"))

        if self._setter(self.element_uid, value):
            self.value = value



import json
import pathlib

from zeroconf import ServiceInfo

file = pathlib.Path(__file__).parent / ".." / "test_data.json"
with file.open("r") as fh:
    test_data = json.load(fh)


class Zeroconf:

    def get_service_info(self, service_type, name):
        service_info = ServiceInfo(service_type, name)
        service_info.server = "devolo-homecontrol.local"
        return service_info



import pytest


@pytest.fixture()
def mock_updater_binary_switch(mocker):
    """ Mock processing an update for binary switch devices. """
    mocker.patch("devolo_home_control_api.publisher.updater.Updater._binary_switch", return_value=None)


@pytest.fixture()
def mock_updater_pending_operations(mocker):
    """ Mock processing an update for a pending operation attribute. """
    mocker.patch("devolo_home_control_api.publisher.updater.Updater._pending_operations", return_value=None)



import pytest

from devolo_home_control_api.homecontrol import HomeControl

from ..mocks.mock_homecontrol import mock__inspect_devices
from ..mocks.mock_mprm_rest import mock_get_data_from_uid_list


@pytest.fixture()
def home_control_instance(request,
                          mydevolo,
                          mock_gateway,
                          mock_mprmwebsocket_websocket_connection,
                          mock_inspect_devices_metering_plug,
                          mock_mprmrest_get_all_zones,
                          mock_mprm__detect_gateway_in_lan,
                          mock_mprmwebsocket_get_remote_session):
    """ Create a mocked Home Control instance with static test data. """
    request.cls.homecontrol = HomeControl(request.cls.gateway.get("id"), mydevolo_instance=mydevolo)


@pytest.fixture()
def mock_inspect_devices_metering_plug(mocker, request, mock_mydevolo__call, mock_mprmrest_get_all_devices):
    """ Create a mocked Home Control setup with a Metering Plug. """
    if request.node.name not in ["test__inspect_devices"]:
        mocker.patch("devolo_home_control_api.homecontrol.HomeControl._inspect_devices", mock__inspect_devices)


@pytest.fixture()
def mock_extract_data_from_element_uids(mocker, request):
    mocker.patch("devolo_home_control_api.backend.mprm_rest.MprmRest.get_data_from_uid_list", mock_get_data_from_uid_list)


@pytest.fixture()
def mock_inspect_devices(mocker):
    mocker.patch("devolo_home_control_api.homecontrol.HomeControl._inspect_devices", return_value=None)



from devolo_home_control_api.backend.mprm_websocket import MprmWebsocket


class StubMprmWebsocket(MprmWebsocket):

    def __init__(self):
        super().__init__()
        self._ws = None
        self._connected = True
        self._reachable = True
        self._event_sequence = 0
        self._url = "https://test.test"

    def detect_gateway_in_lan(self):
        pass

    def get_local_session(self):
        # We are abusing this to raise the expected exception
        raise self._ws()

    def get_remote_session(self):
        pass

    def on_update(self, message):
        pass

    def _post(self, data):
        pass



import pytest

from devolo_home_control_api.helper.string import camel_case_to_snake_case
from devolo_home_control_api.helper.uid import (get_device_type_from_element_uid,
                                                get_device_uid_from_element_uid,
                                                get_device_uid_from_setting_uid,
                                                get_home_id_from_device_uid,
                                                get_sub_device_uid_from_element_uid)


class TestHelper:

    def test_camel_case_to_snake_case(self):
        assert camel_case_to_snake_case("CamelCase") == "camel_case"
        assert camel_case_to_snake_case("camelCase") == "camel_case"

    def test_get_device_type_from_element_uid(self):
        assert get_device_type_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2#2") == "devolo.Meter"

    @pytest.mark.parametrize("element_uid",
                             ["devolo.Meter:hdm:ZWave:F6BF9812/2#2",
                              "devolo.Meter:hdm:ZWave:F6BF9812/2:secure#2"])
    def test_get_device_uid_from_element_uid(self, element_uid):
        assert get_device_uid_from_element_uid(element_uid) == "hdm:ZWave:F6BF9812/2"

    @pytest.mark.parametrize("setting_uid", ["lis.hdm:ZWave:EB5A9F6C/2", "lis.hdm:ZWave:EB5A9F6C/2:secure"])
    def test_get_device_uid_from_setting_uid(self, setting_uid):
        assert get_device_uid_from_setting_uid(setting_uid) == "hdm:ZWave:EB5A9F6C/2"

    def test_get_sub_device_uid_from_element_uid(self):
        assert get_sub_device_uid_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2#2") == 2
        assert get_sub_device_uid_from_element_uid("devolo.Meter:hdm:ZWave:F6BF9812/2") is None

    def test_get_home_id_from_device_uid(self):
        assert get_home_id_from_device_uid("hdm:ZWave:F6BF9812/2") == "F6BF9812"



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.multi_level_switch_property import MultiLevelSwitchProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def multi_level_switch_device(device_uid: str) -> Zwave:
    """
    Represent a multi level switch device in tests

    :param device_uid: Device UID this mock shall have
    :return: Multi level switch device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data.get("devices").get("multi_level_switch"))

    device.multi_level_switch_property = {}
    device.settings_property = {}

    device.multi_level_switch_property[f'devolo.MultiLevelSwitch:{device_uid}'] = MultiLevelSwitchProperty(
        element_uid=f"devolo.MultiLevelSwitch:{device_uid}",
        setter=lambda uid,
        state: None,
        value=test_data.get("devices").get("multi_level_switch").get("value"),
        max=test_data.get("devices").get("multi_level_switch").get("max"),
        min=test_data.get("devices").get("multi_level_switch").get("min"))

    device.settings_property['general_device_settings'] = SettingsProperty(
        element_uid=f"gds.{device_uid}",
        setter=lambda uid,
        state: None,
        icon=test_data.get("devices").get("multi_level_switch").get("icon"),
        name=test_data.get("devices").get("multi_level_switch").get("itemName"),
        zone_id=test_data.get("devices").get("multi_level_switch").get("zoneId"),
        zones=test_data.get("gateway").get("zones"))

    return device



import pytest
from devolo_home_control_api.mydevolo import GatewayOfflineError, Mydevolo, WrongCredentialsError, WrongUrlError


class TestMydevolo:

    def test_credentials_valid(self, mydevolo):
        assert mydevolo.credentials_valid()

    @pytest.mark.usefixtures("mock_mydevolo_uuid_raise_WrongCredentialsError")
    def test_credentials_invalid(self, mydevolo):
        assert not mydevolo.credentials_valid()

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_gateway_ids(self, mydevolo):
        assert mydevolo.get_gateway_ids() == [self.gateway['id']]

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_gateway_ids_empty(self, mydevolo):
        with pytest.raises(IndexError):
            mydevolo.get_gateway_ids()

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_get_full_url(self, mydevolo):
        full_url = mydevolo.get_full_url(self.gateway['id'])
        assert full_url == self.gateway['full_url']

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_get_gateway(self, mydevolo):
        details = mydevolo.get_gateway(self.gateway['id'])
        assert details.get("gatewayId") == self.gateway['id']

    @pytest.mark.usefixtures("mock_mydevolo__call_raise_WrongUrlError")
    def test_get_gateway_invalid(self, mydevolo):
        with pytest.raises(WrongUrlError):
            mydevolo.get_gateway(self.gateway['id'])

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_get_zwave_products(self, mydevolo):
        device_info = mydevolo.get_zwave_products(manufacturer="0x0060", product_type="0x0001", product="0x000")
        assert device_info.get("name") == "Everspring PIR Sensor SP814"

    @pytest.mark.usefixtures("mock_mydevolo__call_raise_WrongUrlError")
    def test_get_zwave_products_invalid(self, mydevolo):
        device_info = mydevolo.get_zwave_products(manufacturer="0x0070", product_type="0x0001", product="0x000")
        assert device_info.get("name") == "Unknown"

    @pytest.mark.usefixtures("mock_mydevolo__call")
    @pytest.mark.parametrize("result", [True, False])
    def test_maintenance(self, mydevolo, result):
        assert mydevolo.maintenance() == result

    def test_set_password(self, mydevolo):
        mydevolo._gateway_ids = [self.gateway['id']]
        mydevolo.password = self.user['password']

        assert mydevolo._uuid is None
        assert mydevolo._gateway_ids == []

    def test_set_user(self, mydevolo):
        mydevolo._gateway_ids = [self.gateway['id']]
        mydevolo.user = self.user['username']

        assert mydevolo._uuid is None
        assert mydevolo._gateway_ids == []

    def test_get_user(self, mydevolo):
        mydevolo.user = self.user['username']
        assert mydevolo.user == self.user['username']

    def test_get_password(self, mydevolo):
        mydevolo.password = self.user['password']
        assert mydevolo.password == self.user['password']

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_uuid(self, mydevolo):
        mydevolo._uuid = None
        assert mydevolo.uuid() == self.user['uuid']

    @pytest.mark.usefixtures("mock_response_wrong_credentials_error")
    def test_call_WrongCredentialsError(self):
        mydevolo = Mydevolo()
        with pytest.raises(WrongCredentialsError):
            mydevolo._call("test")

    @pytest.mark.usefixtures("mock_response_wrong_url_error")
    def test_call_WrongUrlError(self):
        mydevolo = Mydevolo()
        with pytest.raises(WrongUrlError):
            mydevolo._call("test")

    @pytest.mark.usefixtures("mock_response_gateway_offline")
    def test_call_GatewayOfflineError(self, mydevolo):
        with pytest.raises(GatewayOfflineError):
            mydevolo._call("test")

    @pytest.mark.usefixtures("mock_response_valid")
    def test_call_valid(self, mydevolo):
        assert mydevolo._call("test").get("response") == "response"



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.settings_property import SettingsProperty


@pytest.mark.usefixtures("home_control_instance")
class TestSettingsProperty:

    def test_settings_property_valid(self):
        setting_property = SettingsProperty(element_uid=f"lis.{self.devices.get('mains').get('uid')}",
                                            setter=lambda uid,
                                            state: None,
                                            led_setting=True,
                                            events_enabled=False,
                                            param_changed=True,
                                            local_switching=False,
                                            remote_switching=True)
        assert setting_property.led_setting
        assert not setting_property.events_enabled
        assert setting_property.param_changed
        assert not setting_property.local_switching
        assert setting_property.remote_switching

    def test_settings_property_invalid(self):
        with pytest.raises(WrongElementError):
            SettingsProperty(element_uid="invalid", setter=lambda uid, state: None)

    def test__set_bas_valid(self):
        muted = self.devices.get("siren").get("muted")
        self.homecontrol.devices.get(self.devices.get("siren").get("uid")).settings_property.get("muted").set(value=not muted)
        assert self.homecontrol.devices.get(
            self.devices.get("siren").get("uid")).settings_property.get("muted").value is not muted

    def test__set_gds_valid(self):
        self.homecontrol.devices.get(
            self.devices.get("mains").get("uid")).settings_property.get("general_device_settings").set(events_enabled=False)
        assert not self.homecontrol.devices.get(
            self.devices.get("mains").get("uid")).settings_property.get("general_device_settings").events_enabled

    def test__set_lis_valid(self):
        self.homecontrol.devices.get(self.devices.get("mains").get("uid")).settings_property.get("led").set(led_setting=False)
        assert not self.homecontrol.devices.get(self.devices.get("mains").get("uid")).settings_property.get("led").led_setting

    def test__set_mss_invalid(self):
        with pytest.raises(ValueError):
            self.homecontrol.devices.get(
                self.devices.get("sensor").get("uid")).settings_property.get("motion_sensitivity").set(motion_sensitivity=110)

    def test__set_mss_valid(self):
        self.homecontrol.devices.get(
            self.devices.get("sensor").get("uid")).settings_property.get("motion_sensitivity").set(motion_sensitivity=90)
        assert self.homecontrol.devices.get(
            self.devices.get("sensor").get("uid")).settings_property.get("motion_sensitivity").motion_sensitivity == 90

    def test__set_ps_valid(self):
        self.homecontrol.devices.get(
            self.devices.get("mains").get("uid")).settings_property.get("protection").set(local_switching=False)
        assert not self.homecontrol.devices.get(
            self.devices.get("mains").get("uid")).settings_property.get("protection").local_switching

    def test__set_trs_valid(self):
        self.homecontrol.devices.get(
            self.devices.get("sensor").get("uid")).settings_property.get("temperature_report").set(temp_report=False)
        assert not self.homecontrol.devices.get(
            self.devices.get("sensor").get("uid")).settings_property.get("temperature_report").temp_report



import json
import pathlib

import pytest

file = pathlib.Path(__file__).parent / "test_data.json"
with file.open("r") as fh:
    test_data = json.load(fh)

pytest_plugins = [
    'tests.fixtures.gateway',
    'tests.fixtures.homecontrol',
    'tests.fixtures.mprm',
    'tests.fixtures.mydevolo',
    'tests.fixtures.publisher',
    'tests.fixtures.requests',
    'tests.fixtures.socket',
    'tests.fixtures.updater',
]


@pytest.fixture(autouse=True)
def test_data_fixture(request):
    """ Load test data. """
    request.cls.user = test_data['user']
    request.cls.gateway = test_data['gateway']
    request.cls.devices = test_data['devices']



"""Helper functions for splitting UID

Element UIDs and setting UIDs are the central identifies in devolo Home Control. However, for grouping information together
you often need to split them. This helper functions will do that for you.
"""

import re
from typing import Optional


def get_device_uid_from_element_uid(element_uid: str) -> str:
    """
    Return device UID from the given element UID.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
    :return: Device UID, something like hdm:ZWave:CBC56091/24
    """
    parts = re.search(r".*?:(.*/\d{1,3})", element_uid)
    if parts:
        return parts.group(1)
    raise ValueError("Element UID has a wrong format.")


def get_device_uid_from_setting_uid(setting_uid: str) -> str:
    """
    Return the device uid of the given setting UID.

    :param setting_uid: Setting UID, something like lis.hdm:ZWave:EB5A9F6C/2
    :return: Device UID, something like hdm:ZWave:EB5A9F6C/2
    """
    parts = re.search(r".*\.(.*/\d{1,3})", setting_uid)
    if parts:
        return parts.group(1)
    raise ValueError("Settings UID has a wrong format.")


def get_sub_device_uid_from_element_uid(element_uid: str) -> Optional[int]:
    """
    Return the sub device uid of the given element UID.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
    :return: Sub device UID, something like 2
    """
    return None if "#" not in element_uid else int(element_uid.split("#")[-1])


def get_device_type_from_element_uid(element_uid: str) -> str:
    """
    Return the device type of the given element UID.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
    :return: Device type, something like devolo.MultiLevelSensor
    """
    return element_uid.split(":")[0]


def get_home_id_from_device_uid(device_uid: str) -> str:
    """
    Return the home id of the given device UID
    :param device_uid: Device UID, something like hdm:ZWave:EB5A9F6C/4
    :return: Home ID, something like EB5A9F6C
    """
    parts = re.search(r":.*?(:)(.*)(/)", device_uid)
    if parts:
        return parts.group(2)
    raise ValueError("Device UID has a wrong format.")



import logging
from abc import ABC
from datetime import datetime

from ..devices.zwave import get_device_uid_from_element_uid


class Property(ABC):
    """
    Abstract base object for properties.

    :param element_uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24#2
    """

    def __init__(self, element_uid: str):
        self.element_uid = element_uid
        self.device_uid = get_device_uid_from_element_uid(element_uid)

        self._logger = logging.getLogger(self.__class__.__name__)
        self._last_activity = datetime.fromtimestamp(0)  # Set last activity to 1.1.1970. Will be corrected on update.

    @property
    def last_activity(self) -> datetime:
        """ Date and time the property was last updated. """
        return self._last_activity



import pytest
from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty


class TestZwave:

    @pytest.mark.usefixtures("home_control_instance")
    @pytest.mark.usefixtures("mock_mprmrest__extract_data_from_element_uid")
    @pytest.mark.usefixtures("mock_get_zwave_products")
    def test_get_property(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['mains']['properties'])

        device.binary_switch_property = {}
        element_uid = f"devolo.BinarySwitch:{self.devices['mains']['uid']}"
        device.binary_switch_property[element_uid] = BinarySwitchProperty(
            element_uid=element_uid,
            setter=lambda uid,
            state: None,
            state=self.devices['mains']['properties']['state'],
            enabled=self.devices['mains']['properties']['guiEnabled'])

        assert isinstance(device.get_property("binary_switch")[0], BinarySwitchProperty)

    @pytest.mark.usefixtures("mock_mydevolo__call")
    @pytest.mark.usefixtures("mock_get_zwave_products")
    def test_get_property_invalid(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['mains']['properties'])
        with pytest.raises(AttributeError):
            device.get_property("binary_switch")

    @pytest.mark.usefixtures("mock_mydevolo__call")
    @pytest.mark.usefixtures("mock_get_zwave_products")
    def test_battery_level(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['remote'])
        assert device.battery_level == self.devices['remote']['batteryLevel']

    @pytest.mark.usefixtures("mock_mydevolo__call")
    @pytest.mark.usefixtures("mock_get_zwave_products")
    def test_device_online_state_state(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['ambiguous_2'])
        assert device.status == 1

        device = Zwave(mydevolo_instance=mydevolo, **self.devices['mains']['properties'])
        assert device.status == 2

        device = Zwave(mydevolo_instance=mydevolo, **self.devices['ambiguous_1'])
        assert device.status not in [1, 2]

    @pytest.mark.usefixtures("mock_get_zwave_products")
    def test_in_online_online(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['mains']['properties'])
        assert device.is_online()

    @pytest.mark.usefixtures("mock_get_zwave_products")
    def test_in_online_offline(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['offline'])
        assert not device.is_online()

    @pytest.mark.usefixtures("mock_mydevolo__call")
    def test_get_zwave_info(self, mydevolo):
        device = Zwave(mydevolo_instance=mydevolo, **self.devices['mains']['properties'])
        device.get_zwave_info()
        assert device.brand == self.devices['mains']['brand']



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.multi_level_sensor_property import MultiLevelSensorProperty


@pytest.mark.usefixtures("home_control_instance")
class TestMultiLevelSensorProperty:

    def test_multi_level_property_invalid(self):
        with pytest.raises(WrongElementError):
            MultiLevelSensorProperty(element_uid="invalid", setter=lambda uid, state: None)

    def test_unit(self):
        mlsp = MultiLevelSensorProperty(element_uid=self.devices.get("sensor").get("elementUIDs")[2],
                                        setter=lambda uid,
                                        state: None,
                                        sensor_type=self.devices.get("sensor").get("sensor_type"),
                                        value=self.devices.get("sensor").get("value"),
                                        unit=self.devices.get("sensor").get("unit"))
        assert mlsp.unit == "%"



from datetime import datetime
from typing import Union

from ..exceptions.device import WrongElementError
from .property import Property


class ConsumptionProperty(Property):
    """
    Object for consumptions. It stores the current and total consumption and the corresponding units.

    :param element_uid: Element UID, something like devolo.Meter:hdm:ZWave:CBC56091/24#2
    :key current: Consumption value valid at time of creating the instance
    :type current: float
    :key total: Total consumption since last reset
    :type total: float
    :key total_since: Timestamp in milliseconds of last reset
    :type total_since: int
    """

    def __init__(self, element_uid: str, **kwargs: Union[int, float]):
        if not element_uid.startswith("devolo.Meter:"):
            raise WrongElementError(f"{element_uid} is not a Meter.")

        super().__init__(element_uid=element_uid)

        self._current: float = kwargs.pop("current", 0.0)
        self.current_unit = "W"
        self._total: float = kwargs.pop("total", 0.0)
        self.total_unit = "kWh"
        self._total_since = datetime.utcfromtimestamp(kwargs.pop("total_since", 0) / 1000)

    @property
    def current(self) -> float:
        """ Consumption value. """
        return self._current

    @current.setter
    def current(self, current: float):
        """ Update current consumption and set point in time of the last_activity. """
        self._current = current
        self._last_activity = datetime.now()
        self._logger.debug("current of element_uid %s set to %s.", self.element_uid, current)

    @property
    def total(self) -> float:
        """ Total consumption value. """
        return self._total

    @total.setter
    def total(self, total: float):
        """ Update total consumption and set point in time of the last_activity. """
        self._total = total
        self._last_activity = datetime.now()
        self._logger.debug("total of element_uid %s set to %s.", self.element_uid, total)

    @property
    def total_since(self) -> datetime:
        """ Date and time the binary sensor was last triggered. """
        return self._total_since

    @total_since.setter
    def total_since(self, timestamp: int):
        """ Convert a timestamp in millisecond to a datetime object. """
        self._total_since = datetime.utcfromtimestamp(timestamp / 1000)
        self._logger.debug("total_since of element_uid %s set to %s.", self.element_uid, self._total_since)



import pytest

from ..mocks.mock_response import (MockResponseConnectTimeout,
                                   MockResponseGet,
                                   MockResponseJsonError,
                                   MockResponsePost,
                                   MockResponseReadTimeout,
                                   MockResponseServiceUnavailable)


@pytest.fixture()
def mock_response_gateway_offline(mocker):
    """ Mock requests get method with service_unavailable status_code. """
    mocker.patch("requests.get",
                 return_value=MockResponseServiceUnavailable({"link": "test_link"},
                                                             status_code=503))


@pytest.fixture()
def mock_response_json(mocker):
    """ Mock requests.Session get method with success status_code. """
    mocker.patch("requests.Session",
                 return_value=MockResponseGet({"link": "test_link"},
                                              status_code=200))


@pytest.fixture()
def mock_response_requests_ConnectTimeout(mocker):
    """ Mock requests.Session with ConnectTimeout exception. """
    mocker.patch("requests.Session",
                 return_value=MockResponseConnectTimeout({"link": "test_link"},
                                                         status_code=200))


@pytest.fixture()
def mock_response_json_JSONDecodeError(mocker):
    """ Mock requests.Session with JSONDecodeError exception. """
    mocker.patch("requests.Session",
                 return_value=MockResponseJsonError({"link": "test_link"},
                                                    status_code=200))


@pytest.fixture()
def mock_response_requests_invalid_id(mocker):
    """ Mock requests.Session with JSONDecodeError exception. """
    mocker.patch("requests.Session",
                 return_value=MockResponsePost({"link": "test_link"},
                                               status_code=200))


@pytest.fixture()
def mock_response_requests_valid(mocker):
    """ Mock requests.Session post method with success status_code. """
    mocker.patch("requests.Session",
                 return_value=MockResponsePost({"link": "test_link"},
                                               status_code=200))


@pytest.fixture()
def mock_response_valid(mocker):
    """ Mock requests get method with success status_code. """
    mocker.patch("requests.get",
                 return_value=MockResponseGet({"response": "response"},
                                              status_code=200))


@pytest.fixture()
def mock_response_wrong_credentials_error(mocker):
    """ Mock requests get method with forbidden status_code. """
    mocker.patch("requests.get",
                 return_value=MockResponseGet({"link": "test_link"},
                                              status_code=403))


@pytest.fixture()
def mock_response_wrong_url_error(mocker):
    """ Mock requests get method with notfound status_code. """
    mocker.patch("requests.get",
                 return_value=MockResponseGet({"link": "test_link"},
                                              status_code=404))


@pytest.fixture()
def mock_response_requests_ReadTimeout(mocker):
    """ Mock requests.Session with ReadTimeout exception. """
    mocker.patch("requests.Session",
                 return_value=MockResponseReadTimeout({"link": "test_link"},
                                                      status_code=200))


@pytest.fixture()
def mock_session_get(mocker, request):
    """ Mock requests.Session get method with test data. """
    properties = {
        'test_get_local_session_valid': {
            'link': 'test_link'
        },
        'test__on_pong': {
            'link': 'test_link'
        },
    }
    mocker.patch("requests.Session", return_value=MockResponseGet(properties.get(request.node.name), status_code=200))


@pytest.fixture()
def mock_session_get_nok(mocker):
    mocker.patch("requests.Session",
                 return_value=MockResponseServiceUnavailable({},
                                                             status_code=500))



import pytest
from devolo_home_control_api.mydevolo import Mydevolo, WrongCredentialsError, WrongUrlError

from ..mocks.mock_mydevolo import MockMydevolo


@pytest.fixture()
def mydevolo(request):
    """ Create real mydevolo object with static test data. """
    mydevolo_instance = Mydevolo()
    mydevolo_instance._uuid = request.cls.user.get("uuid")
    yield mydevolo_instance


@pytest.fixture()
def mock_mydevolo__call(mocker, request):
    """ Mock calls to the mydevolo API. """
    mock_mydevolo = MockMydevolo(request)
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo._call", side_effect=mock_mydevolo._call)


@pytest.fixture()
def mock_mydevolo__call_raise_WrongUrlError(mocker):
    """ Respond with WrongUrlError on calls to the mydevolo API. """
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo._call", side_effect=WrongUrlError)


@pytest.fixture()
def mock_mydevolo_uuid_raise_WrongCredentialsError(mocker):
    """ Respond with WrongCredentialsError on calls to the mydevolo API. """
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo.uuid", side_effect=WrongCredentialsError)


@pytest.fixture()
def mock_get_zwave_products(mocker):
    """ Mock Z-Wave product information call to speed up tests. """
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo.get_zwave_products",
                 return_value={})



import json
import logging
import sys
from abc import ABC

from requests import Session
from requests.exceptions import ConnectionError, ReadTimeout  # pylint: disable=redefined-builtin

from ..devices.gateway import Gateway
from ..exceptions.gateway import GatewayOfflineError
from ..mydevolo import Mydevolo


class MprmRest(ABC):
    """
    The abstract MprmRest object handles calls to the so called mPRM. It does not cover all API calls, just those requested
    up to now. All calls are done in a gateway context, so you have to create a derived class, that provides a Gateway object
    and a Session object.
    """

    def __init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._data_id = 0
        self._local_ip = ""
        self._url = ""

        self._mydevolo: Mydevolo
        self._session: Session
        self.gateway: Gateway

    def get_all_devices(self) -> list:
        """
        Get all devices.

        :return: All devices and their properties.
        """
        self._logger.info("Inspecting devices")
        data = {
            "method": "FIM/getFunctionalItems",
            "params": [['devolo.DevicesPage'],
                       0]
        }
        response = self._post(data)
        self._logger.debug("Response of 'get_all_devices':\n%s", response)
        return response['result']['items'][0]['properties']['deviceUIDs']

    def get_all_zones(self) -> dict:
        """
        Get all zones, also called rooms.

        :return: All zone IDs and their name.
        """
        self._logger.debug("Inspecting zones")
        data = {
            "method": "FIM/getFunctionalItems",
            "params": [["devolo.Grouping"],
                       0]
        }
        response = self._post(data)['result']['items'][0]['properties']['zones']
        self._logger.debug("Response of 'get_all_zones':\n%s", response)
        return {key['id']: key['name']
                for key in response}

    def get_data_from_uid_list(self, uids: list) -> list:
        """
        Returns data from an element UID list using an RPC call.

        :param uids: Element UIDs, something like [devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2,
                     devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#1]
        :return: Data connected to the element UIDs, payload so to say
        """
        data = {
            "method": "FIM/getFunctionalItems",
            "params": [uids,
                       0]
        }
        response = self._post(data)
        self._logger.debug("Response of 'get_data_from_uid_list':\n%s", response)
        return response['result']['items']

    def get_name_and_element_uids(self, uid: str):
        """
        Returns the name, all element UIDs and the device model of the given device UID.

        :param uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
        """
        data = {
            "method": "FIM/getFunctionalItems",
            "params": [[uid],
                       0]
        }
        response = self._post(data)
        self._logger.debug("Response of 'get_name_and_element_uids':\n%s", response)
        return response['result']['items'][0]['properties']

    def refresh_session(self):
        """
        Refresh currently running session. Without this call from time to time especially websockets will terminate.
        """
        self._logger.debug("Refreshing session.")
        data = {
            "method": "FIM/invokeOperation",
            "params": [f"devolo.UserPrefs.{self._mydevolo.uuid()}",
                       "resetSessionTimeout",
                       []]
        }
        self._post(data)

    def set_binary_switch(self, uid: str, state: bool) -> bool:
        """
        Set a binary switch state of a device.

        :param uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24
        :param state: True if switching on, False if switching off
        :return: True if successfully switched, false otherwise
        """
        data = {
            "method": "FIM/invokeOperation",
            "params": [uid,
                       "turnOn" if state else "turnOff",
                       []]
        }
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=state, response=response)

    def set_multi_level_switch(self, uid: str, value: float) -> bool:
        """
        Set a multi level switch value of a device.

        :param uid: Element UID, something like devolo.Dimmer:hdm:ZWave:CBC56091/24
        :param value: Value the multi level switch shall have
        :return: True if successfully switched, false otherwise
        """
        data = {
            "method": "FIM/invokeOperation",
            "params": [uid,
                       "sendValue",
                       [value]]
        }
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=value, response=response)

    def set_remote_control(self, uid: str, key_pressed: int) -> bool:
        """
        Press the button of a remote control virtually.

        :param uid: Element UID, something like devolo.RemoteControl:hdm:ZWave:CBC56091/24
        :param key_pressed: Number of the button pressed
        :return: True if successfully switched, false otherwise
        """
        data = {
            "method": "FIM/invokeOperation",
            "params": [uid,
                       "pressKey",
                       [key_pressed]]
        }
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=key_pressed, response=response)

    def set_setting(self, uid: str, setting: list) -> bool:
        """
        Set a setting of a device.

        :param uid: Element UID, something like acs.hdm:ZWave:CBC56091/24
        :param setting: Settings to set
        :return: True if successfully switched, false otherwise
        """
        data = {
            "method": "FIM/invokeOperation",
            "params": [uid,
                       "save",
                       setting]
        }
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=setting, response=response)

    def _evaluate_response(self, uid, value, response):
        """ Evaluate the response of setting a device to a value. """
        if response['result'].get("status") == 1:
            return True
        if response['result'].get("status") == 2:
            self._logger.debug("Value of %s is already %s.", uid, value)
        else:
            self._logger.error("Something went wrong setting %s.", uid)
            self._logger.debug("Response to set command:\n%s", response)
        return False

    def _post(self, data: dict) -> dict:
        """
        Communicate with the RPC interface. If the call times out, it is assumed that the gateway is offline and the state is
        changed accordingly.

        :param data: Data to be send
        :return: Response to the data
        """
        self._data_id += 1
        data['jsonrpc'] = "2.0"
        data['id'] = self._data_id
        try:
            response = self._session.post(self._url + "/remote/json-rpc",
                                          data=json.dumps(data),
                                          headers={
                                              "content-type": "application/json"
                                          },
                                          timeout=30).json()
        except (ConnectionError, ReadTimeout):
            self._logger.error("Gateway is offline.")
            self._logger.debug(sys.exc_info())
            self.gateway.update_state(False)
            raise GatewayOfflineError("Gateway is offline.") from None
        if response['id'] != data['id']:
            self._logger.error("Got an unexpected response after posting data.")
            self._logger.debug("Message had ID %s, response had ID %s.", data['id'], response['id'])
            raise ValueError("Got an unexpected response after posting data.")
        return response



import contextlib
import socket
import sys
import time
from abc import ABC
from json import JSONDecodeError
from threading import Thread
from typing import Optional
from urllib.parse import urlsplit

import requests
from zeroconf import ServiceBrowser, ServiceStateChange, Zeroconf

from ..exceptions.gateway import GatewayOfflineError
from .mprm_websocket import MprmWebsocket


class Mprm(MprmWebsocket, ABC):
    """
    The abstract Mprm object handles the connection to the devolo Cloud (remote) or the gateway in your LAN (local). Either
    way is chosen, depending on detecting the gateway via mDNS.
    """

    def __init__(self):
        self._zeroconf: Optional[Zeroconf]

        super().__init__()

        self.detect_gateway_in_lan()
        self.create_connection()

    def create_connection(self):
        """
        Create session, either locally or remotely via cloud. The remote case has two conditions, that both need to be
        fulfilled: Remote access must be allowed and my devolo must not be in maintenance mode.
        """
        if self._local_ip:
            self.gateway.local_connection = True
            self.get_local_session()
        elif self.gateway.external_access and not self._mydevolo.maintenance():
            self.get_remote_session()
        else:
            self._logger.error("Cannot connect to gateway. No gateway found in LAN and external access is not possible.")
            raise ConnectionError("Cannot connect to gateway.")

    def detect_gateway_in_lan(self) -> str:
        """
        Detect a gateway in local network via mDNS and check if it is the desired one. Unfortunately, the only way to tell is
        to try a connection with the known credentials. If the gateway is not found within 3 seconds, it is assumed that a
        remote connection is needed.

        :return: Local IP of the gateway, if found
        """
        zeroconf = self._zeroconf or Zeroconf()
        browser = ServiceBrowser(zeroconf, "_http._tcp.local.", handlers=[self._on_service_state_change])
        self._logger.info("Searching for gateway in LAN.")
        start_time = time.time()
        while not time.time() > start_time + 3 and self._local_ip == "":
            time.sleep(0.05)

        Thread(target=browser.cancel, name=f"{self.__class__.__name__}.browser_cancel").start()
        if not self._zeroconf:
            Thread(target=zeroconf.close, name=f"{self.__class__.__name__}.zeroconf_close").start()

        return self._local_ip

    def get_local_session(self) -> bool:
        """
        Connect to the gateway locally. Calling a special portal URL on the gateway returns a second URL with a token. Calling
        that URL establishes the connection.
        """
        self._logger.info("Connecting to gateway locally.")
        self._url = "http://" + self._local_ip
        self._logger.debug("Session URL set to '%s'", self._url)
        try:
            connection = self._session.get(self._url + "/dhlp/portal/full",
                                           auth=(self.gateway.local_user,
                                                 self.gateway.local_passkey),
                                           timeout=5)
        except (requests.exceptions.ConnectTimeout, requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
            self._logger.error("Could not connect to the gateway locally.")
            self._logger.debug(sys.exc_info())
            raise GatewayOfflineError("Gateway is offline.") from None

        # After a reboot we can connect to the gateway but it answers with a 503 if not fully started.
        if not connection.ok:
            self._logger.error("Could not connect to the gateway locally.")
            self._logger.debug("Gateway start-up is not finished, yet.")
            raise GatewayOfflineError("Gateway is offline.") from None

        token_url = connection.json()['link']
        self._logger.debug("Got a token URL: %s", token_url)

        self._session.get(token_url)
        return True

    def get_remote_session(self) -> bool:
        """
        Connect to the gateway remotely. Calling the known portal URL is enough in this case.
        """
        self._logger.info("Connecting to gateway via cloud.")
        try:
            url = urlsplit(self._session.get(self.gateway.full_url, timeout=15).url)
            self._url = f"{url.scheme}://{url.netloc}"
            self._logger.debug("Session URL set to '%s'", self._url)
        except JSONDecodeError:
            self._logger.error("Could not connect to the gateway remotely.")
            self._logger.debug(sys.exc_info())
            raise GatewayOfflineError("Gateway is offline.") from None
        return True

    def _on_service_state_change(self, zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange):
        """ Service handler for Zeroconf state changes. """
        if state_change is ServiceStateChange.Added:
            service_info = zeroconf.get_service_info(service_type, name)
            if service_info and service_info.server.startswith("devolo-homecontrol"):
                with contextlib.suppress(requests.exceptions.ReadTimeout), \
                     contextlib.suppress(requests.exceptions.ConnectTimeout):
                    self._try_local_connection(service_info.addresses)

    def _try_local_connection(self, addresses: list):
        """ Try to connect to an mDNS hostname. If connection was successful, save local IP address. """
        for address in addresses:
            ip = socket.inet_ntoa(address)
            if requests.get("http://" + ip + "/dhlp/port/full",
                            auth=(self.gateway.local_user,
                                  self.gateway.local_passkey),
                            timeout=0.5).status_code == requests.codes.ok:  # pylint: disable=no-member
                self._logger.debug("Got successful answer from ip %s. Setting this as local gateway", ip)
                self._local_ip = ip



import logging
from typing import Dict

from ..mydevolo import Mydevolo


class Gateway:
    """
    Representing object for devolo Home Control Central Units. As it is a gateway from the IP world to the Z-Wave
    world, we call it that way. Nearly all attributes are delivered by my devolo.

    :param gateway_id: Gateway ID (aka serial number), typically found on the label of the device
    :param mydevolo_instance: Mydevolo instance for talking to the devolo Cloud
    """

    def __init__(self, gateway_id: str, mydevolo_instance: Mydevolo):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._mydevolo = mydevolo_instance

        details = self._mydevolo.get_gateway(gateway_id)

        self.id = details['gatewayId']
        self.name = details.get("name")
        self.role = details.get("role")
        self.full_url = self._mydevolo.get_full_url(self.id)
        self.local_user = self._mydevolo.uuid()
        self.local_passkey = details['localPasskey']
        self.external_access = details.get("externalAccess")
        self.firmware_version = details.get("firmwareVersion")

        # Let's assume the gateway is connected remotely. Will be corrected as soon as a real connection is established.
        self.local_connection = False

        # Let's assume the gateway is offline. Will be corrected the twinkling of an eye.
        self.online = False
        self.sync = False

        self.zones: Dict = {}
        self.home_id = ""

        self._update_state(status=details.get("status", ""), state=details.get("state", ""))

    def update_state(self, online: bool = None):
        """
        Update the state of the gateway. If called without parameter, we will check my devolo.

        :param online: Detected state of the gateway
        """
        if online is None:
            details = self._mydevolo.get_gateway(self.id)
            self._update_state(status=details.get("status", ""), state=details.get("state", ""))
        else:
            self.online = online
            self.sync = online

    def _update_state(self, status: str, state: str):
        """ Helper to update the state. """
        self.online = status == "devolo.hc_gateway.status.online"
        self.sync = state == "devolo.hc_gateway.state.idle"



import logging

import requests

from . import __version__
from .exceptions.gateway import GatewayOfflineError
from .exceptions.general import WrongCredentialsError, WrongUrlError


class Mydevolo:
    """
    The Mydevolo object handles calls to the my devolo API v1. It does not cover all API calls, just those requested up to now.
    All calls are done in a user context, so you need to provide credentials of that user.
    """

    def __init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._user = None
        self._password = None
        self._uuid = None
        self._gateway_ids = []

        self.url = "https://www.mydevolo.com"

    @property
    def user(self) -> str:
        """ The user (also known as my devolo ID) is used for basic authentication. """
        return self._user

    @user.setter
    def user(self, user: str):
        """ Invalidate uuid and gateway IDs on user name change. """
        self._user = user
        self._uuid = None
        self._gateway_ids = []

    @property
    def password(self) -> str:
        """ The password is used for basic authentication. """
        return self._password

    @password.setter
    def password(self, password: str):
        """ Invalidate uuid and gateway IDs on password change. """
        self._password = password
        self._uuid = None
        self._gateway_ids = []

    def credentials_valid(self) -> bool:
        """
        Check if current credentials are valid. This is done by trying to get the UUID. If that fails, credentials must be
        wrong. If it succeeds, we can reuse the UUID for later usages.
        """
        try:
            self.uuid()
            return True
        except WrongCredentialsError:
            return False

    def get_gateway_ids(self) -> list:
        """
        Get all gateway IDs attached to current account.
        """
        if not self._gateway_ids:
            self._logger.debug("Getting list of gateways")
            items = self._call(f"{self.url}/v1/users/{self.uuid()}/hc/gateways/status")['items']
            for gateway in items:
                self._gateway_ids.append(gateway.get("gatewayId"))
                self._logger.debug("Adding %s to list of gateways.", gateway.get("gatewayId"))
            if len(self._gateway_ids) == 0:
                self._logger.error("Could not get gateway list. No gateway attached to account?")
                raise IndexError("No gateways found.")
        return self._gateway_ids

    def get_gateway(self, gateway_id: str) -> dict:
        """
        Get gateway details like name, local passkey and other.

        :param gateway_id: Gateway ID
        :return: Gateway details
        """
        self._logger.debug("Getting details for gateway %s", gateway_id)
        try:
            details = self._call(f"{self.url}/v1/users/{self.uuid()}/hc/gateways/{gateway_id}")
        except WrongUrlError:
            self._logger.error("Could not get full URL. Wrong gateway ID used?")
            raise
        return details

    def get_full_url(self, gateway_id: str) -> str:
        """
        Get gateway's portal URL.

        :param gateway_id: Gateway ID
        :return: URL to access the gateway's portal.
        """
        self._logger.debug("Getting full URL of gateway.")
        return self._call(f"{self.url}/v1/users/{self.uuid()}/hc/gateways/{gateway_id}/fullURL")['url']

    def get_zwave_products(self, manufacturer: str, product_type: str, product: str) -> dict:
        """
        Get information about a Z-Wave device.

        :param manufacturer: The manufacturer ID in hex.
        :param product_type: The product type ID in hex.
        :param product: The product ID in hex.
        :return: All known product information.
        """
        self._logger.debug("Getting information for %s/%s/%s", manufacturer, product_type, product)
        try:
            device_info = self._call(f"{self.url}/v1/zwave/products/{manufacturer}/{product_type}/{product}")
        except WrongUrlError:
            # At some devices no device information are returned
            self._logger.debug("No device info found")
            device_info = {
                "brand": "devolo" if manufacturer == "0x0175" else "Unknown",
                "deviceType": "Unknown",
                "genericDeviceClass": "Unknown",
                "href": f"{self.url}/v1/zwave/products/{manufacturer}/{product_type}/{product}",
                "identifier": "Unknown",
                "isZWavePlus": False,
                "manufacturerId": manufacturer,
                "name": "Unknown",
                "productId": product,
                "productTypeId": product_type,
                "specificDeviceClass": "Unknown",
                "zwaveVersion": "Unknown",
            }
        return device_info

    def maintenance(self) -> bool:
        """
        If devolo Home Control is in maintenance, there is not much we can do via cloud.
        """
        state = self._call(f"{self.url}/v1/hc/maintenance")['state']
        if state == "on":
            return False
        self._logger.warning("devolo Home Control is in maintenance mode.")
        return True

    def uuid(self) -> str:
        """
        The uuid is a central attribute in my devolo. Most URLs in the user's context contain it.
        """
        if self._uuid is None:
            self._logger.debug("Getting UUID")
            self._uuid = self._call(f"{self.url.rstrip('/')}/v1/users/uuid")['uuid']
        return self._uuid

    def _call(self, url: str) -> dict:
        """ Make a call to any entry point with the user's context. """
        headers = {
            "content-type": "application/json",
            "User-Agent": f"devolo_home_control_api/{__version__}"
        }
        responds = requests.get(url, auth=(self._user, self._password), headers=headers, timeout=60)

        if responds.status_code == requests.codes.forbidden:  # pylint: disable=no-member
            self._logger.error("Could not get full URL. Wrong username or password?")
            raise WrongCredentialsError("Wrong username or password.")
        if responds.status_code == requests.codes.not_found:  # pylint: disable=no-member
            raise WrongUrlError(f"Wrong URL: {url}")
        if responds.status_code == requests.codes.service_unavailable:  # pylint: disable=no-member
            # mydevolo sends a 503, if the gateway is offline
            self._logger.error("The requested gateway seems to be offline.")
            raise GatewayOfflineError("Gateway offline.")

        return responds.json()



from datetime import datetime
from typing import Callable

from ..exceptions.device import WrongElementError
from .property import Property


class RemoteControlProperty(Property):
    """
    Object for remote controls. It stores the button state and additional information that help displaying the state in the
    right context.

    :param element_uid: Element UID, something like devolo.RemoteControl:hdm:ZWave:CBC56091/24#2
    :param setter: Method to call on setting the state
    :key key_count: Number of buttons this remote control has
    :type key_count: int
    :key key_pressed: Number of the button pressed
    :type key_pressed: int
    """

    def __init__(self, element_uid: str, setter: Callable, **kwargs: int):
        if not element_uid.startswith("devolo.RemoteControl"):
            raise WrongElementError(f"{element_uid} is not a remote control.")

        super().__init__(element_uid=element_uid)
        self._setter = setter

        self._key_pressed: int = kwargs.pop("key_pressed", 0)
        self.key_count: int = kwargs.pop("key_count", 0)

    @property
    def key_pressed(self) -> int:
        """ Multi level value. """
        return self._key_pressed

    @key_pressed.setter
    def key_pressed(self, key_pressed: int):
        """ Update value of the multilevel value and set point in time of the last_activity. """
        self._key_pressed = key_pressed
        self._last_activity = datetime.now()
        self._logger.debug("key_pressed of element_uid %s set to %s.", self.element_uid, key_pressed)

    def set(self, key_pressed: int):
        """
        Trigger a button press of a remote control like if the button was physically pressed.

        :param key_pressed: Number of the button to press
        """
        if key_pressed > self.key_count or key_pressed <= 0:
            raise ValueError(f"Set value {key_pressed} is invalid.")

        if self._setter(self.element_uid, key_pressed):
            self.key_pressed = key_pressed
            self._logger.debug("Remote Control property %s set to %s", self.element_uid, key_pressed)



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.binary_sensor_property import BinarySensorProperty
from devolo_home_control_api.properties.humidity_bar_property import HumidityBarProperty
from devolo_home_control_api.properties.multi_level_sensor_property import MultiLevelSensorProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def humidity_sensor_device(device_uid: str) -> Zwave:
    """
    Represent a humidity sensor

    :param device_uid: Device UID this mock shall have
    :return: Humidity sensor device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data.get("devices").get("humidity"))

    device.binary_sensor_property = {}
    device.humidity_bar_property = {}
    device.multi_level_sensor_property = {}
    device.settings_property = {}

    element_uid = f"devolo.DewpointSensor:{device_uid}"
    device.multi_level_sensor_property[element_uid] = MultiLevelSensorProperty(
        element_uid=element_uid,
        value=test_data.get("devices").get("humidity").get("dewpoint"))

    element_uid = f"devolo.MildewSensor:{device_uid}"
    device.binary_sensor_property[element_uid] = BinarySensorProperty(
        element_uid=element_uid,
        state=test_data.get("devices").get("humidity").get("mildew"))

    element_uid = f"devolo.HumidityBar:{device_uid}"
    device.humidity_bar_property[element_uid] = HumidityBarProperty(
        element_uid=element_uid,
        zone=test_data.get("devices").get("humidity").get("zone"),
        value=test_data.get("devices").get("humidity").get("value"))

    device.settings_property['general_device_settings'] = SettingsProperty(
        element_uid=f'gds.{test_data.get("devices").get("humidity").get("uid")}',
        setter=lambda uid,
        state: None,
        icon=test_data.get("devices").get("humidity").get("icon"),
        name=test_data.get("devices").get("humidity").get("itemName"),
        zone_id=test_data.get("devices").get("humidity").get("zoneId"),
        zones=test_data.get("gateway").get("zones"))

    return device



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.multi_level_switch_property import MultiLevelSwitchProperty


@pytest.mark.usefixtures("home_control_instance")
class TestMultiLevelSwitchProperty:

    def test_multi_level_switch_property_invalid(self):
        with pytest.raises(WrongElementError):
            MultiLevelSwitchProperty(element_uid="invalid", setter=lambda uid, state: None)

    def test_unit(self):
        element_uid = self.devices['siren']['elementUIDs'][0]
        switch_type = self.devices['siren']['switch_type']
        mlsp = MultiLevelSwitchProperty(element_uid=element_uid, setter=lambda uid, state: None, switch_type=switch_type)
        assert mlsp.unit is None

    def test_set_invalid(self):
        value = self.devices['multi_level_switch']['max'] + 1
        element_uid = self.devices['multi_level_switch']['elementUIDs'][0]
        device = self.homecontrol.devices.get(self.devices['multi_level_switch']['uid'])
        device.multi_level_switch_property[element_uid]._setter = lambda uid, state: True
        with pytest.raises(ValueError):
            device.multi_level_switch_property[element_uid].set(value=value)

    def test_set(self):
        value = self.devices['multi_level_switch']['value']
        element_uid = self.devices['multi_level_switch']['elementUIDs'][0]
        device = self.homecontrol.devices.get(self.devices['multi_level_switch']['uid'])
        device.multi_level_switch_property[element_uid]._setter = lambda uid, state: True
        device.multi_level_switch_property[element_uid].set(value=value)
        assert device.multi_level_switch_property[element_uid].value == value



import logging
from typing import Callable, Dict, KeysView, Union


class Publisher:
    """
    The Publisher send messages to attached subscribers.
    """

    def __init__(self, events: Union[list, KeysView]):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._events: Dict = {event: {}
                              for event in events}

    def add_event(self, event: str):
        """ Add a new event to listen to. """
        self._events[event] = {}

    def delete_event(self, event: str):
        """ Delete a not longer needed event. """
        self._events.pop(event)

    def dispatch(self, event: str, message: tuple):
        """ Dispatch the message to the subscribers. """
        for callback in self._get_subscribers_for_specific_event(event).values():
            callback(message)

    def register(self, event: str, who: object, callback: Callable = None):
        """
        As a new subscriber for an event, add a callback function to call on new message.
        If no callback is given, it registers update().

        :raises AttributeError: The supposed callback is not callable.
        """
        if callback is None:
            callback = getattr(who, "update")
        self._get_subscribers_for_specific_event(event)[who] = callback
        self._logger.debug("Subscriber registered for event %s", event)

    def unregister(self, event: str, who: object):
        """ Remove a subscriber for a specific event. """
        del self._get_subscribers_for_specific_event(event)[who]
        self._logger.debug("Subscriber deleted for event %s", event)

    def _get_subscribers_for_specific_event(self, event: str) -> Dict:
        """ All subscribers listening to an event. """
        return self._events.get(event,
                                {})



from datetime import datetime
from typing import Callable

from ..exceptions.device import SwitchingProtected, WrongElementError
from .property import Property


class BinarySwitchProperty(Property):
    """
    Object for binary switches. It stores the binary switch state.

    :param element_uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24#2
    :param setter: Method to call on setting the state
    :key enabled: State of the remote protection setting
    :type enabled: bool
    :key state: State the switch has at time of creating this instance
    :type state: bool
    """

    def __init__(self, element_uid: str, setter: Callable, **kwargs: bool):
        if not element_uid.startswith("devolo.BinarySwitch:"):
            raise WrongElementError(f"{element_uid} is not a Binary Switch.")

        super().__init__(element_uid=element_uid)
        self._setter = setter

        self._state: bool = kwargs.pop("state", False)
        self.enabled: bool = kwargs.pop("enabled", False)

    @property
    def state(self) -> bool:
        """ State of the binary sensor. """
        return self._state

    @state.setter
    def state(self, state: bool):
        """ Update state of the binary sensor and set point in time of the last_activity. """
        self._state = state
        self._last_activity = datetime.now()
        self._logger.debug("State of %s set to %s.", self.element_uid, state)

    def set(self, state: bool):
        """
        Set the binary switch of the given element_uid to the given state.

        :param state: True if switching on, False if switching off
        """
        if not self.enabled:
            raise SwitchingProtected("This device is protected against remote switching.")

        if self._setter(self.element_uid, state):
            self.state = state



import pytest
from devolo_home_control_api.exceptions.device import SwitchingProtected, WrongElementError
from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty


@pytest.mark.usefixtures("home_control_instance")
class TestBinarySwitchProperty:

    def test_binary_switch_property_invalid(self):
        with pytest.raises(WrongElementError):
            BinarySwitchProperty(element_uid="invalid", setter=lambda uid, state: None, state=True, enabled=True)

    def test_set(self):
        uid = self.devices['mains']['uid']
        element_uid = self.devices['mains']['elementUIDs'][1]
        self.homecontrol.devices[uid].binary_switch_property[element_uid]._setter = lambda uid, state: True
        self.homecontrol.devices[uid].binary_switch_property[element_uid].set(True)
        assert self.homecontrol.devices[uid].binary_switch_property[element_uid].state

    def test_set_protected(self):
        uid = self.devices['mains']['uid']
        element_uid = self.devices['mains']['elementUIDs'][1]
        self.homecontrol.devices[uid].binary_switch_property[element_uid]._setter = lambda uid, state: True
        self.homecontrol.devices[uid].binary_switch_property[element_uid].enabled = False
        with pytest.raises(SwitchingProtected):
            self.homecontrol.devices[uid].binary_switch_property[element_uid].set(True)



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.multi_level_switch_property import MultiLevelSwitchProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def siren(device_uid: str) -> Zwave:
    """
    Represent a siren in tests

    :param device_uid: Device UID this mock shall have
    :return: Siren device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data['devices']['siren'])

    device.multi_level_switch_property = {}
    device.multi_level_switch_property[f'devolo.SirenMultiLevelSwitch:{device_uid}'] = MultiLevelSwitchProperty(
        element_uid=f"devolo.SirenMultiLevelSwitch:{device_uid}",
        setter=lambda uid,
        state: None,
        state=test_data['devices']['siren']['state'])

    device.settings_property = {}
    device.settings_property['muted'] = SettingsProperty(element_uid=f"bas.{device_uid}",
                                                         setter=lambda uid,
                                                         state: True,
                                                         value=test_data['devices']['siren']['muted'])

    device.settings_property["general_device_settings"] = SettingsProperty(element_uid=f"gds.{device_uid}",
                                                                           setter=lambda uid,
                                                                           state: None,
                                                                           icon=test_data['devices']['siren']['icon'],
                                                                           name=test_data['devices']['siren']['itemName'],
                                                                           zone_id=test_data['devices']['siren']['zoneId'],
                                                                           zones=test_data['gateway']['zones'])

    device.settings_property["tone"] = SettingsProperty(element_uid=f"mss.{device_uid}",
                                                        setter=lambda uid,
                                                        state: None,
                                                        value=test_data['devices']['siren']['properties']['value'])

    return device



import pytest
import requests

from devolo_home_control_api.backend.mprm_rest import MprmRest

from ..mocks.mock_gateway import MockGateway
from ..mocks.mock_mprm_rest import try_local_connection
from ..mocks.mock_service_browser import ServiceBrowser
from ..mocks.mock_websocket import try_reconnect
from ..mocks.mock_websocketapp import MockWebsocketapp
from ..mocks.mock_zeroconf import Zeroconf
from ..stubs.mprm import StubMprm
from ..stubs.mprm_websocket import StubMprmWebsocket


@pytest.fixture()
def mock_mprm_get_local_session(mocker):
    """ Mock getting a local session to speed up tests. """
    mocker.patch("devolo_home_control_api.backend.mprm.Mprm.get_local_session", return_value=True)


@pytest.fixture()
def mock_mprm_service_browser(mocker):
    """ Mock Zeroconf ServiceBrowser. """
    mocker.patch("zeroconf.ServiceBrowser.__init__", ServiceBrowser.__init__)
    mocker.patch("zeroconf.ServiceBrowser.cancel", ServiceBrowser.cancel)
    mocker.patch("zeroconf.Zeroconf.get_service_info", Zeroconf.get_service_info)


@pytest.fixture()
def mock_mprm__detect_gateway_in_lan(mocker, request):
    """ Mock detecting a gateway in the local area network to speed up tests. """
    if request.node.name not in ["test_detect_gateway_in_lan_valid", "test_detect_gateway_in_lan"]:
        mocker.patch("devolo_home_control_api.backend.mprm.Mprm.detect_gateway_in_lan",
                     return_value=request.cls.gateway['local_ip'])


@pytest.fixture()
def mock_mprm__try_local_connection(mocker):
    """ Mock finding gateway's IP. """
    mocker.patch("devolo_home_control_api.backend.mprm.Mprm._try_local_connection", try_local_connection)


@pytest.fixture()
def mock_mprmrest_get_all_devices(mocker, request):
    """ Mock getting all devices from the mPRM. """
    if request.node.name not in ["test_get_all_devices"]:
        mocker.patch("devolo_home_control_api.backend.mprm_rest.MprmRest.get_all_devices",
                     return_value=["hdm:ZWave:F6BF9812/2"])


@pytest.fixture()
def mock_mprmrest_get_all_zones(mocker, request):
    """ Mock getting all zones from the mPRM. """
    if request.node.name not in ["test_get_all_zones"]:
        mocker.patch("devolo_home_control_api.backend.mprm_rest.MprmRest.get_all_zones",
                     return_value=request.cls.gateway["zones"])


@pytest.fixture()
def mock_mprmrest__extract_data_from_element_uid(mocker, request):
    """ Mock extracting device data. """
    properties = {
        "test_fetch_binary_switch_state_valid_on": {
            "properties": {
                "state": 1
            }
        },
        "test_fetch_binary_switch_state_valid_off": {
            "properties": {
                "state": 0
            }
        },
        "test_fetch_consumption_valid": {
            "properties": {
                "currentValue": request.cls.devices.get("mains").get("properties").get("current_consumption"),
                "totalValue": request.cls.devices.get("mains").get("properties").get("total_consumption"),
            }
        },
        "test_fetch_led_setting_valid": {
            "properties": {
                "led": request.cls.devices.get("mains").get("properties").get("led_setting")
            }
        },
        "test_fetch_param_changed_valid": {
            "properties": {
                "paramChanged": request.cls.devices.get("mains").get("properties").get("param_changed")
            }
        },
        "test_fetch_general_device_settings_valid": {
            "properties": {
                "eventsEnabled": request.cls.devices.get("mains").get("properties").get("events_enabled"),
                "name": request.cls.devices.get("mains").get("properties").get("name"),
                "icon": request.cls.devices.get("mains").get("properties").get("icon"),
                "zoneID": request.cls.devices.get("mains").get("properties").get("zone_id")
            }
        },
        "test_fetch_protection_setting_valid": {
            "properties": {
                "localSwitch": request.cls.devices.get("mains").get("properties").get("local_switch"),
                "remoteSwitch": request.cls.devices.get("mains").get("properties").get("remote_switch")
            }
        },
        "test_fetch_voltage_valid": {
            "properties": {
                "value": request.cls.devices.get("mains").get("properties").get("voltage")
            }
        },
        "test_update_consumption_valid": {
            "properties": {
                "currentValue": request.cls.devices.get("mains").get("properties").get("current_consumption"),
                "totalValue": request.cls.devices.get("mains").get("properties").get("total_consumption")
            }
        },
        "test__inspect_device": None,
    }

    mocker.patch("devolo_home_control_api.backend.mprm_rest.MprmRest.get_data_from_uid_list",
                 return_value=properties.get(request.node.name))


@pytest.fixture()
def mock_mprmrest__post(mocker, request):
    """ Mock getting properties from the mPRM. """
    test_case = request.node.name.split('[')[0]
    properties = {
        "test_get_name_and_element_uids": {
            "result": {
                "items": [{
                    "properties": {
                        "itemName": "test_name",
                        "zone": "test_zone",
                        "batteryLevel": "test_battery",
                        "icon": "test_icon",
                        "elementUIDs": "test_element_uids",
                        "settingUIDs": "test_setting_uids",
                        "deviceModelUID": "test_device_model_uid",
                        "status": "test_status",
                    }
                }]
            }
        },
        "test_get_data_from_uid_list": {
            "result": {
                "items": [{
                    "properties": {
                        "itemName": "test_name"
                    }
                }]
            }
        },
        "test_get_all_devices": {
            "result": {
                "items": [{
                    "properties": {
                        "deviceUIDs": "deviceUIDs"
                    }
                }]
            }
        },
        "test_get_all_zones": {
            "result": {
                "items": [{
                    "properties": {
                        "zones": [{
                            "id": "hz_3",
                            "name": "Office"
                        }]
                    }
                }]
            }
        },
        "test_set_success": {
            "result": {
                "status": 1
            }
        },
        "test_set_failed": {
            "result": {
                "status": 0
            }
        },
        "test_set_doubled": {
            "result": {
                "status": 2
            }
        },
    }

    mocker.patch("devolo_home_control_api.backend.mprm_rest.MprmRest._post", return_value=properties[test_case])


@pytest.fixture()
def mock_mprmwebsocket_get_remote_session(mocker, request):
    """ Mock getting a remote session to speed up tests. """
    if request.node.name not in ["test_get_remote_session_JSONDecodeError"]:
        mocker.patch("devolo_home_control_api.backend.mprm.Mprm.get_remote_session", return_value=True)


@pytest.fixture()
def mock_mprmwebsocket_on_update(mocker):
    """ Mock websocket message processing. """
    mocker.patch("devolo_home_control_api.backend.mprm_websocket.MprmWebsocket.on_update", return_value=None)


@pytest.fixture()
def mock_mprmwebsocket_try_reconnect(mocker):
    """ Mock reconnect attemt as successful. """
    mocker.patch("devolo_home_control_api.backend.mprm_websocket.MprmWebsocket._try_reconnect", try_reconnect)


@pytest.fixture()
def mock_mprmwebsocket_websocketapp(mocker):
    """ Mock a websocket connection init. """
    mocker.patch("websocket.WebSocketApp.__init__", MockWebsocketapp.__init__)
    mocker.patch("websocket.WebSocketApp.close", MockWebsocketapp.close)
    mocker.patch("websocket.WebSocketApp.run_forever", MockWebsocketapp.run_forever)


@pytest.fixture()
def mock_mprmwebsocket_websocket_connection(mocker):
    """ Mock a running websocket connection to speed up tests. """
    mocker.patch("devolo_home_control_api.backend.mprm_websocket.MprmWebsocket.wait_for_websocket_establishment",
                 return_value=False)
    mocker.patch("devolo_home_control_api.backend.mprm_websocket.MprmWebsocket.websocket_connect", return_value=None)


@pytest.fixture()
def mock_mprmwebsocket_websocket_disconnect(mocker):
    """ Mock closing a websocket connection. """
    mocker.patch("devolo_home_control_api.backend.mprm_websocket.MprmWebsocket.websocket_disconnect", return_value=None)


@pytest.fixture()
def mprm_instance(request,
                  mocker,
                  mydevolo,
                  mock_gateway,
                  mock_inspect_devices_metering_plug,
                  mock_mprm__detect_gateway_in_lan):
    """ Create a mocked mPRM instance with static test data. """
    if "TestMprmRest" in request.node.nodeid:
        request.cls.mprm = MprmRest()
    elif "TestMprmWebsocket" in request.node.nodeid:
        request.cls.mprm = StubMprmWebsocket()
    else:
        mocker.patch("devolo_home_control_api.backend.mprm_websocket.MprmWebsocket.websocket_connect", return_value=None)
        request.cls.mprm = StubMprm()
        request.cls.mprm.gateway = MockGateway(request.cls.gateway.get("id"), mydevolo=mydevolo)


@pytest.fixture()
def mprm_session():
    """ Mock a valid mPRM session. """
    session = requests.Session()
    session.url = "https://homecontrol.mydevolo.com"
    return session



import pytest
from devolo_home_control_api.devices.gateway import Gateway


@pytest.mark.usefixtures("mydevolo")
@pytest.mark.usefixtures("mock_mydevolo__call")
class TestGateway:

    def test_update_state_known(self, mydevolo):
        gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
        gateway.update_state(False)

        assert not gateway.online
        assert not gateway.sync

    def test_update_state_offline(self, mydevolo):
        gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
        gateway._update_state(status="devolo.hc_gateway.status.offline", state="devolo.hc_gateway.state.update")

        assert not gateway.online
        assert not gateway.sync

    def test_update_state_unknown(self, mydevolo):
        gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
        gateway.online = False
        gateway.sync = False
        gateway.update_state()

        assert gateway.online
        assert gateway.sync



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.humidity_bar_property import HumidityBarProperty


@pytest.mark.usefixtures("home_control_instance")
class TestHumidityBar:

    def test_humidity_bar_property_invalid(self):
        with pytest.raises(WrongElementError):
            HumidityBarProperty(element_uid="invalid")



import pytest
import zeroconf

from devolo_home_control_api.exceptions.gateway import GatewayOfflineError


@pytest.mark.usefixtures("mprm_instance")
class TestMprm:

    @pytest.mark.usefixtures("mock_mprm_get_local_session")
    def test_create_connection_local(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway["local_ip"]
        self.mprm.create_connection()

    @pytest.mark.usefixtures("mock_mprmwebsocket_get_remote_session")
    @pytest.mark.usefixtures("mock_session_get")
    def test_create_connection_remote(self, mprm_session, mydevolo):
        self.mprm._mydevolo = mydevolo
        self.mprm._session = mprm_session
        self.mprm.gateway.external_access = True
        self.mprm.create_connection()

    def test_create_connection_invalid(self):
        with pytest.raises(ConnectionError):
            self.mprm.gateway.external_access = False
            self.mprm.create_connection()

    @pytest.mark.usefixtures("mock_mprm_service_browser")
    def test_detect_gateway_in_lan(self):
        self.mprm._local_ip = self.gateway["local_ip"]
        assert self.mprm.detect_gateway_in_lan() == self.gateway["local_ip"]

    @pytest.mark.usefixtures("mock_session_get")
    @pytest.mark.usefixtures("mock_response_json")
    def test_get_local_session_valid(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway["local_ip"]
        assert self.mprm.get_local_session()

    @pytest.mark.usefixtures("mock_response_requests_ConnectTimeout")
    def test_get_local_session_ConnectTimeout(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._local_ip = self.gateway["local_ip"]
        with pytest.raises(GatewayOfflineError):
            self.mprm.get_local_session()

    @pytest.mark.usefixtures("mock_session_get_nok")
    def test_get_local_session_connection_nok(self, mprm_session):
        self.mprm._session = mprm_session
        with pytest.raises(GatewayOfflineError):
            self.mprm.get_local_session()

    @pytest.mark.usefixtures("mock_response_json_JSONDecodeError")
    def test_get_remote_session_JSONDecodeError(self, mprm_session):
        self.mprm._session = mprm_session
        with pytest.raises(GatewayOfflineError):
            self.mprm.get_remote_session()

    @pytest.mark.usefixtures("mock_mprm_service_browser")
    @pytest.mark.usefixtures("mock_mprm__try_local_connection")
    def test__on_service_state_change(self):
        zc = zeroconf.Zeroconf()
        service_type = "_http._tcp.local."
        self.mprm._on_service_state_change(zc, service_type, service_type, zeroconf.ServiceStateChange.Added)
        assert self.mprm._local_ip == self.gateway["local_ip"]

    @pytest.mark.usefixtures("mock_socket_inet_ntoa")
    @pytest.mark.usefixtures("mock_response_valid")
    def test__try_local_connection_success(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._try_local_connection([self.gateway["local_ip"]])
        assert self.mprm._local_ip == self.gateway["local_ip"]



import pytest

from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.binary_sensor_property import BinarySensorProperty


@pytest.mark.usefixtures("home_control_instance")
class TestBinarySensorProperty:

    def test_binary_sensor_property_invalid(self):
        with pytest.raises(WrongElementError):
            BinarySensorProperty(element_uid="invalid", state=True)

    def test_get_binary_sensor(self):
        assert self.homecontrol.devices.get(self.devices.get("sensor").get("uid")).binary_sensor_property.get(
            self.devices.get("sensor").get("elementUIDs")[0]).state



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.binary_sensor_property import BinarySensorProperty
from devolo_home_control_api.properties.multi_level_sensor_property import MultiLevelSensorProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def multi_level_sensor_device(device_uid: str) -> Zwave:
    """
    Represent a Multi Level Sensor

    :param device_uid: Device UID this mock shall have
    :return: Multi Level Sensor device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data.get("devices").get("sensor"))

    device.binary_sensor_property = {}
    device.multi_level_sensor_property = {}
    device.settings_property = {}

    element_uid = f"devolo.BinarySensor:{device_uid}"
    device.binary_sensor_property[element_uid] = BinarySensorProperty(
        element_uid=element_uid,
        state=test_data.get("devices").get("sensor").get("state"))

    element_uid = f"devolo.MultiLevelSensor:{device_uid}#MultilevelSensor(1)"
    device.multi_level_sensor_property[element_uid] = MultiLevelSensorProperty(
        element_uid=element_uid,
        value=test_data.get("devices").get("sensor").get("value"),
        unit=test_data.get("devices").get("sensor").get("unit"))

    device.settings_property['temperature_report'] = SettingsProperty(
        element_uid=f"trs.{device_uid}",
        setter=lambda uid,
        state: True,
        temp_report=test_data.get("devices").get("sensor").get("temp_report"))
    device.settings_property['motion_sensitivity'] = SettingsProperty(
        element_uid=f"mss.{device_uid}",
        setter=lambda uid,
        state: True,
        motion_sensitivity=test_data.get("devices").get("sensor").get("motion_sensitivity"))
    device.settings_property["general_device_settings"] = SettingsProperty(
        element_uid=f'gds.{device_uid}',
        setter=lambda uid,
        state: None,
        icon=test_data.get("devices").get("sensor").get("icon"),
        name=test_data.get("devices").get("sensor").get("itemName"),
        zone_id=test_data.get("devices").get("sensor").get("zoneId"),
        zones=test_data.get("gateway").get("zones"))

    return device



from datetime import datetime

from ..exceptions.device import WrongElementError
from .sensor_property import SensorProperty


class MultiLevelSensorProperty(SensorProperty):
    """
    Object for multi level sensors. It stores the multi level sensor state and additional information that help displaying the
    state in the right context.

    :param element_uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#MultilevelSensor(1)
    :key value: Multi level value
    :type value: float
    :key unit: Unit of that value
    :type unit: int
    """

    def __init__(self, element_uid: str, **kwargs):
        if not element_uid.startswith(("devolo.DewpointSensor:",
                                       "devolo.MultiLevelSensor:",
                                       "devolo.ValveTemperatureSensor",
                                       "devolo.VoltageMultiLevelSensor:")):
            raise WrongElementError(f"{element_uid} is not a Multi Level Sensor.")

        super().__init__(element_uid=element_uid, **kwargs)

        self._value: float = kwargs.pop("value", 0.0)
        self._unit: int = kwargs.pop("unit", 0)

    @property
    def unit(self) -> str:
        """ Human readable unit of the property. """
        units = {
            "dewpoint": {
                0: "°C",
                1: "°F"
            },
            "humidity": {
                0: "%",
                1: "g/m³"
            },
            "light": {
                0: "%",
                1: "lx"
            },
            "temperature": {
                0: "°C",
                1: "°F"
            },
            "Seismic Intensity": {
                0: ""
            },
            "voltage": {
                0: "V",
                1: "mV"
            },
        }
        try:
            return units[self.sensor_type].get(self._unit, str(self._unit))
        except KeyError:
            return str(self._unit)

    @property
    def value(self) -> float:
        """ Multi level value. """
        return self._value

    @value.setter
    def value(self, value: float):
        """ Update value of the multilevel sensor and set point in time of the last_activity. """
        self._value = value
        self._last_activity = datetime.now()
        self._logger.debug("value of element_uid %s set to %s.", self.element_uid, value)



"""Helper functions for converting strings

The information we get from the gateway is often returned in a way, that does not fit to our naming convention. This helper
function make names fit in nicely.
"""

import re


def camel_case_to_snake_case(expression: str) -> str:
    """
    Turn CamelCaseStrings to snake_case_strings. This is used where the original Java names should by more pythonic.

    :param: expression: Expression, that should be converted to snake case
    :return: Expression in snake case
    """
    return re.sub("([a-z])([A-Z]+)", r"\1_\2", expression).lower()



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def dummy_device(key: str) -> Zwave:
    """
    Represent a dummy device in tests, where we do not care about the device type

    :param key: Key to look up in test_data.json
    :return: Dummy device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data['devices'][key])

    device.binary_switch_property = {}
    device.binary_switch_property[f"devolo.BinarySwitch:{test_data['devices'][key]['uid']}"] = BinarySwitchProperty(
        element_uid=f"devolo.BinarySwitch:{test_data['devices'][key]['uid']}",
        setter=lambda uid,
        state: None,
        state=test_data['devices'][key]['state'],
        enabled=test_data['devices'][key]['guiEnabled'])

    device.settings_property = {}
    device.settings_property["general_device_settings"] = SettingsProperty(
        element_uid=f"gds.{test_data['devices'][key]['uid']}",
        setter=lambda uid,
        state: None,
        icon=test_data['devices'][key]['icon'],
        name=test_data['devices'][key]['itemName'],
        zone_id=test_data['devices'][key]['zoneId'],
        zones=test_data.get("gateway").get("zones"))

    return device



class MprmDeviceCommunicationError(Exception):
    """ Communicating to a device via mPRM failed. """


class MprmDeviceNotFoundError(Exception):
    """ A device like this was not found. """


class WrongElementError(Exception):
    """ This element was not meant for this property. """


class SwitchingProtected(Exception):
    """ This device is protected against remote switching. """



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.remote_control_property import RemoteControlProperty


@pytest.mark.usefixtures("home_control_instance")
class TestRemoteControlProperty:

    def test_remote_control_property_invalid(self):
        with pytest.raises(WrongElementError):
            RemoteControlProperty(element_uid="invalid", setter=lambda uid, state: None)

    def test_set_invalid(self):
        with pytest.raises(ValueError):
            self.homecontrol.devices.get(self.devices.get("remote").get("uid"))\
                .remote_control_property.get(self.devices.get("remote").get("elementUIDs")[0]).set(5)

    def test_set_valid(self):
        self.homecontrol.devices.get(self.devices.get("remote").get("uid"))\
            .remote_control_property.get(self.devices.get("remote").get("elementUIDs")[0]).set(1)
        assert self.homecontrol.devices.get(self.devices.get("remote").get("uid"))\
            .remote_control_property.get(self.devices.get("remote").get("elementUIDs")[0]).key_pressed



import json
import threading
import time
from abc import ABC, abstractmethod

import requests
import websocket
from pkg_resources import parse_version
from urllib3.connection import ConnectTimeoutError

from ..exceptions.gateway import GatewayOfflineError
from .mprm_rest import MprmRest


class MprmWebsocket(MprmRest, ABC):
    """
    The abstract MprmWebsocket object handles calls to the mPRM via websockets. It does not cover all API calls, just those
    requested up to now. All calls are done in a gateway context, so you have to create a derived class, that provides a
    Gateway object and a Session object. Further, the derived class needs to implement methods to connect to the websocket,
    either local or remote. Last but not least, the derived class needs to implement a method that is called on new messages.

    The websocket connection itself runs in a thread, that might not terminate as expected. Using a with-statement is
    recommended.
    """

    def __init__(self):
        super().__init__()
        self._ws: websocket.WebSocketApp = None
        self._connected = False  # This attribute saves, if the websocket is fully established
        self._reachable = True  # This attribute saves, if the a new session can be established
        self._event_sequence = 0

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.websocket_disconnect()

    @abstractmethod
    def detect_gateway_in_lan(self):
        pass

    @abstractmethod
    def get_local_session(self):
        pass

    @abstractmethod
    def get_remote_session(self):
        pass

    @abstractmethod
    def on_update(self, message):
        pass

    def wait_for_websocket_establishment(self):
        """
        In some cases it is needed to wait for the websocket to be fully established. This method can be used to block your
        current thread for up to one minute.
        """
        start_time = time.time()
        while not self._connected and time.time() < start_time + 600:
            time.sleep(0.1)
        if not self._connected:
            self._logger.debug("Websocket could not be established")
            raise GatewayOfflineError("Websocket could not be established.")

    def websocket_connect(self):
        """
        Set up the websocket connection. The protocol type of the known session URL is exchanged depending on whether TLS is
        used or not. After establishing the websocket, a ping is sent every 30 seconds to keep the connection alive. If there
        is no response within 5 seconds, the connection is terminated with error state.
        """
        ws_url = self._url.replace("https://", "wss://").replace("http://", "ws://")
        cookie = "; ".join(str(name) + "=" + str(value) for name, value in self._session.cookies.items())

        ws_url = f"{ws_url}/remote/events/?topics=com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED," \
                 f"com/prosyst/mbs/services/fim/FunctionalItemEvent/UNREGISTERED" \
                 f"&filter=(|(GW_ID={self.gateway.id})(!(GW_ID=*)))"
        self._logger.debug("Connecting to %s", ws_url)
        if parse_version(websocket.__version__) < parse_version("0.58.0"):  # Just in case we hit an outdated version
            self._logger.warning(
                "Please consider updating your websocket-client version. Support for <0.58.0 will be removed soon.")
            self._ws = websocket.WebSocketApp(ws_url,
                                              cookie=cookie,
                                              on_open=self._on_open_old,
                                              on_message=self._on_message_old,
                                              on_error=self._on_error_old,
                                              on_close=self._on_close,
                                              on_pong=self._on_pong)
        else:
            self._ws = websocket.WebSocketApp(ws_url,
                                              cookie=cookie,
                                              on_open=self._on_open,
                                              on_message=self._on_message,
                                              on_error=self._on_error,
                                              on_close=self._on_close,
                                              on_pong=self._on_pong)
        self._ws.run_forever(ping_interval=30, ping_timeout=5)

    def websocket_disconnect(self, event: str = ""):
        """
        Close the websocket connection.
        """
        self._logger.info("Closing web socket connection.")
        if event:
            self._logger.info("Reason: %s", event)
        self._ws.close()

    def _on_close(self, *args):  # pylint: disable=unused-argument
        """ Callback method to react on closing the websocket. """
        self._logger.info("Closed websocket connection.")

    def _on_error(self, ws: websocket.WebSocketApp, error: Exception):
        """ Callback method to react on errors. We will try reconnecting with prolonging intervals. """
        self._logger.exception(error)
        self._connected = False
        self._reachable = False
        ws.close()
        self._event_sequence = 0

        sleep_interval = 16
        while not self._reachable:
            self._try_reconnect(sleep_interval)
            sleep_interval = sleep_interval * 2 if sleep_interval < 2048 else 3600

        self.websocket_connect()

    def _on_error_old(self, error: Exception):
        """ Deprecated callback method to react on errors. """
        self._on_error(ws=self._ws, error=error)

    def _on_message(self, ws: websocket.WebSocketApp, message: str):  # pylint: disable=unused-argument
        """ Callback method to react on a message. """
        msg = json.loads(message)
        self._logger.debug("Got message from websocket:\n%s", msg)
        event_sequence = msg["properties"]["com.prosyst.mbs.services.remote.event.sequence.number"]
        if event_sequence == self._event_sequence:
            self._event_sequence += 1
        else:
            self._logger.warning(
                "We missed a websocket message. Internal event_sequence is at %s. "
                "Event sequence by websocket is at %s",
                self._event_sequence,
                event_sequence)
            self._event_sequence = event_sequence + 1
            self._logger.debug("self._event_sequence is set to %s", self._event_sequence)

        self.on_update(msg)

    def _on_message_old(self, message: str):
        """ Deprecated callback method to react on a message. """
        self._on_message(ws=self._ws, message=message)

    def _on_open(self, ws: websocket.WebSocketApp):
        """ Callback method to keep the websocket open. """

        def run():
            self._logger.info("Starting web socket connection.")
            while ws.sock is not None and ws.sock.connected:
                time.sleep(1)

        threading.Thread(target=run, name=f"{self.__class__.__name__}.websocket_run").start()
        self._connected = True

    def _on_open_old(self):
        """ Deprecated callback method to keep the websocket open. """
        self._on_open(ws=self._ws)

    def _on_pong(self, *args):  # pylint: disable=unused-argument
        """ Callback method to keep the session valid. """
        self.refresh_session()

    def _try_reconnect(self, sleep_interval: int):
        """ Try to reconnect to the websocket. """
        try:
            self._logger.info("Trying to reconnect to the websocket.")
            self._reachable = self.get_local_session() if self._local_ip else self.get_remote_session()
        except (ConnectTimeoutError, GatewayOfflineError):
            self._logger.info("Sleeping for %s seconds.", sleep_interval)
            time.sleep(sleep_interval)
        except (requests.exceptions.ConnectTimeout, requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
            self._logger.info("Sleeping for %s seconds.", sleep_interval)
            time.sleep(sleep_interval - 3)  # mDNS browsing will take up tp 3 seconds by itself
            self.detect_gateway_in_lan()



import json
import logging
from typing import Any, Callable, Optional

from ..backend import MESSAGE_TYPES
from ..devices.gateway import Gateway
from ..helper.string import camel_case_to_snake_case
from ..helper.uid import get_device_type_from_element_uid, get_device_uid_from_element_uid, get_device_uid_from_setting_uid
from .publisher import Publisher


class Updater:
    """
    The Updater takes care of new states and values of devices and sends them to the Publisher object. Using methods in here
    do not effect the real device states.

    :param devices: List of devices to await updates for
    :param gateway: Instance of a Gateway object
    :param publisher: Instance of a Publisher object
    """

    def __init__(self, devices: dict, gateway: Gateway, publisher: Publisher):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._gateway = gateway
        self._publisher = publisher

        self.devices = devices
        self.on_device_change: Optional[Callable] = None

    def update(self, message: dict):
        """
        Update states and values depending on the message type.

        :param message: Message to process
        """
        unwanted_properties = [
            ".unregistering",
            "assistantsConnected",
            "operationStatus",
        ]

        # Early return on unwanted messages
        if "UNREGISTERED" in message['topic'] \
                or message['properties']['property.name'] in unwanted_properties \
                or "smartGroup" in message['properties']['uid']:
            return

        # Handle pending operations messages
        if "property.name" in message['properties'] and message['properties']['property.name'] == "pendingOperations":
            self._pending_operations(message)
            return

        # Handle all other messages
        message_type = MESSAGE_TYPES.get(get_device_type_from_element_uid(message['properties']['uid']), "_unknown")
        try:
            getattr(self, message_type)(message)
        except (AttributeError, KeyError):
            # Sometime we receive already messages although the device is not setup yet.
            pass

    def _automatic_calibration(self, message: dict):
        """ Update a automatic calibration message. """
        try:
            calibration_status = message['properties']['property.value.new']['status']
            self._update_automatic_calibration(element_uid=message['properties']['uid'],
                                               calibration_status=calibration_status != 2)
        except (KeyError, TypeError):
            if type(message['properties']['property.value.new']) not in [dict, list]:
                self._update_automatic_calibration(element_uid=message['properties']['uid'],
                                                   calibration_status=bool(message['properties']['property.value.new']))

    def _binary_async(self, message: dict):
        """ Update a binary async setting. """
        if type(message['properties']['property.value.new']) not in [dict, list]:
            element_uid = message['properties']['uid']
            value = bool(message['properties']['property.value.new'])
            device_uid = get_device_uid_from_setting_uid(element_uid)
            try:
                self.devices[device_uid].settings_property[camel_case_to_snake_case(element_uid).split("#")[-1]].value = value
            except KeyError:
                # Siren setting is not initialized like others.
                self.devices[device_uid].settings_property['muted'].value = value
            self._logger.debug("Updating state of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _binary_sync(self, message: dict):
        """ Update a binary sync setting. """
        element_uid = message['properties']['uid']
        value = bool(message['properties']['property.value.new'])
        device_uid = get_device_uid_from_setting_uid(element_uid)
        self.devices[device_uid].settings_property["movement_direction"].direction = value
        self._logger.debug("Updating state of %s to %s", element_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value))

    def _binary_sensor(self, message: dict):
        """ Update a binary sensor's state. """
        if message['properties']['property.value.new'] is not None:
            element_uid = message['properties']['uid']
            value = bool(message['properties']['property.value.new'])
            device_uid = get_device_uid_from_element_uid(element_uid)
            self.devices[device_uid].binary_sensor_property[element_uid].state = value
            self._logger.debug("Updating state of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _binary_switch(self, message: dict):
        """ Update a binary switch's state. """
        if message['properties']['property.name'] == "targetState" \
                and message['properties']['property.value.new'] is not None:
            element_uid = message['properties']['uid']
            value = bool(message['properties']['property.value.new'])
            device_uid = get_device_uid_from_element_uid(element_uid)
            self.devices[device_uid].binary_switch_property[element_uid].state = value
            self._logger.debug("Updating state of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _pending_operations(self, message: dict):
        """ Update pending operation state. """
        element_uid = message['properties']['uid']

        # Early return on useless messages
        if element_uid in [
                "devolo.PairDevice",
                "devolo.RemoveDevice",
                "devolo.mprm.gw.GatewayManager",
        ]:
            return

        pending_operations = bool(message['properties'].get('property.value.new'))
        try:
            device_uid = get_device_uid_from_element_uid(element_uid)
            self.devices[device_uid].pending_operations = pending_operations
        except KeyError:
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self.devices[device_uid].pending_operations = pending_operations
        self._logger.debug("Updating pending operations of device %s to %s", device_uid, pending_operations)
        self._publisher.dispatch(device_uid, ("pending_operations", pending_operations))

    def _current_consumption(self, message: dict):
        """ Update current consumption. """
        self._update_consumption(element_uid=message['uid'], consumption="current", value=message['property.value.new'])

    def _device_state(self, message: dict):
        """ Update the device state. """
        property_name = {
            "batteryLevel": "battery_level",
            "batteryLow": "battery_low",
            "status": "status",
        }

        device_uid = message['properties']['uid']
        name = message['properties']['property.name']
        value = message['properties']['property.value.new']

        try:
            self._logger.debug("Updating %s of %s to %s", property_name[name], device_uid, value)
            setattr(self.devices[device_uid], property_name[name], value)
            self._publisher.dispatch(device_uid, (device_uid, value, property_name[name]))
        except KeyError:
            self._unknown(message)

    def _gateway_accessible(self, message: dict):
        """ Update the gateway's state. """
        if message['properties']['property.name'] == "gatewayAccessible":
            accessible = message['properties']['property.value.new']['accessible']
            online_sync = message['properties']['property.value.new']['onlineSync']
            self._logger.debug("Updating status and state of gateway to status: %s and state: %s", accessible, online_sync)
            self._gateway.online = accessible
            self._gateway.sync = online_sync

    def _general_device(self, message: dict):
        """ Update general device settings. """
        self._update_general_device_settings(element_uid=message['properties']['uid'],
                                             events_enabled=message['properties']['property.value.new']['eventsEnabled'],
                                             icon=message['properties']['property.value.new']['icon'],
                                             name=message['properties']['property.value.new']['name'],
                                             zone_id=message['properties']['property.value.new']['zoneID'],
                                             zones=self._gateway.zones)

    def _grouping(self, message: dict):
        """ Update zone (also called room) of a device. """
        self._gateway.zones = {key['id']: key['name']
                               for key in message['properties']['property.value.new']}
        self._logger.debug("Updating gateway zones.")

    def _gui_enabled(self, message: dict):
        """ Update protection setting of binary switches. """
        device_uid = get_device_uid_from_element_uid(message['uid'])
        enabled = message['property.value.new']
        for element_uid in self.devices[device_uid].binary_switch_property:
            self.devices[device_uid].binary_switch_property[element_uid].enabled = enabled
            self._logger.debug("Updating enabled state of %s to %s", element_uid, enabled)
            self._publisher.dispatch(device_uid, (element_uid, enabled, "gui_enabled"))

    def _humidity_bar(self, message: dict):
        """ Update a humidity bar. """
        fake_element_uid = f"devolo.HumidityBar:{message['properties']['uid'].split(':', 1)[1]}"
        value = message['properties']['property.value.new']
        device_uid = get_device_uid_from_element_uid(fake_element_uid)
        if message['properties']['uid'].startswith("devolo.HumidityBarZone"):
            self.devices[device_uid].humidity_bar_property[fake_element_uid].zone = value
            self._logger.debug("Updating humidity bar zone of %s to %s", fake_element_uid, value)
        elif message['properties']['uid'].startswith("devolo.HumidityBarValue"):
            self.devices[device_uid].humidity_bar_property[fake_element_uid].value = value
            self._logger.debug("Updating humidity bar value of %s to %s", fake_element_uid, value)
        self._publisher.dispatch(device_uid,
                                 (fake_element_uid,
                                  self.devices[device_uid].humidity_bar_property[fake_element_uid].zone,
                                  self.devices[device_uid].humidity_bar_property[fake_element_uid].value))

    def _inspect_devices(self, message: dict):
        """ Call method if a new device appears or an old one disappears. """
        if not callable(self.on_device_change):
            self._logger.error("on_device_change is not set.")
            return

        if not isinstance(message['properties']['property.value.new'], list) \
                or message['properties']['uid'] != "devolo.DevicesPage":
            return

        device_uid, mode = self.on_device_change(device_uids=message['properties']['property.value.new'])
        if mode == "add":
            self._logger.info("%s added.", device_uid)
            self._publisher.add_event(event=device_uid)
            self._publisher.dispatch(device_uid, (device_uid, mode))
        else:
            self._publisher.dispatch(device_uid, (device_uid, mode))
            self._publisher.delete_event(event=device_uid)

    def _led(self, message: dict):
        """ Update LED settings. """
        if type(message['properties']['property.value.new']) not in [dict, list]:
            element_uid = message['properties']['uid']
            value = message['properties']['property.value.new']
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self._logger.debug("Updating %s to %s.", element_uid, value)
            self.devices[device_uid].settings_property['led'].led_setting = value
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _meter(self, message: dict):
        """ Update a meter value. """
        property_name = {
            "currentValue": self._current_consumption,
            "totalValue": self._total_consumption,
            "sinceTime": self._since_time,
            "guiEnabled": self._gui_enabled,
        }

        property_name[message['properties']['property.name']](message['properties'])

    def _multilevel_async(self, message: dict):
        """ Update multilevel async setting (mas) properties. """
        device_uid = get_device_uid_from_setting_uid(message['properties']['uid'])
        try:
            name = camel_case_to_snake_case(message['properties']['itemId'])
        # The Metering Plug has an multilevel async setting without an ID
        except KeyError:
            if self.devices[device_uid].device_model_uid == "devolo.model.Wall:Plug:Switch:and:Meter":
                name = "flash_mode"
            else:
                raise
        self.devices[device_uid].settings_property[name].value = message['properties']['property.value.new']

    def _multi_level_sensor(self, message: dict):
        """ Update a multi level sensor. """
        element_uid = message['properties']['uid']
        value = message['properties']['property.value.new']
        device_uid = get_device_uid_from_element_uid(element_uid)
        self._logger.debug("Updating %s to %s.", element_uid, value)
        self.devices[device_uid].multi_level_sensor_property[element_uid].value = value
        self._publisher.dispatch(device_uid, (element_uid, value))

    def _multi_level_switch(self, message: dict):
        """ Update a multi level switch. """
        if not isinstance(message['properties']['property.value.new'], (list, dict, type(None))):
            element_uid = message['properties']['uid']
            value = message['properties']['property.value.new']
            device_uid = get_device_uid_from_element_uid(element_uid)
            self._logger.debug("Updating %s to %s.", element_uid, value)
            self.devices[device_uid].multi_level_switch_property[element_uid].value = value
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _multilevel_sync(self, message: dict):
        """ Update multilevel sync settings. """
        if type(message['properties']['property.value.new']) not in [dict, list]:
            element_uid = message['properties']['uid']
            value = message['properties']['property.value.new']
            device_uid = get_device_uid_from_setting_uid(element_uid)
            device_model = self.devices[device_uid].device_model_uid
            self._logger.debug("Updating %s to %s.", element_uid, value)
            sync_type = {
                "devolo.model.Siren": "tone",
                "devolo.model.OldShutter": "shutter_duration",
                "devolo.model.Shutter": "shutter_duration",
            }

            try:
                setattr(self.devices[device_uid].settings_property[sync_type[device_model]], sync_type[device_model], value)
            except KeyError:
                # Other devices are up to now always motion sensors.
                self.devices[device_uid].settings_property['motion_sensitivity'].motion_sensitivity = value

            self._publisher.dispatch(device_uid, (element_uid, value))

    def _parameter(self, message: dict):
        """ Update parameter settings. """
        if type(message['properties'].get("property.value.new")) not in [dict, list]:
            element_uid = message['properties']['uid']
            param_changed = message['properties']['property.value.new']
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self.devices[device_uid].settings_property['param_changed'].param_changed = param_changed
            self._logger.debug("Updating %s to %s.", element_uid, param_changed)
            self._publisher.dispatch(device_uid, (element_uid, param_changed))

    def _protection(self, message: dict):
        """ Update protection settings. """
        if type(message['properties'].get("property.value.new")) not in [dict, list]:
            element_uid = message['properties']['uid']
            value = message['properties']['property.value.new']
            name = message['properties']['property.name']
            device_uid = get_device_uid_from_setting_uid(element_uid)
            switching_type = {
                "targetLocalSwitch": "local_switching",
                "localSwitch": "local_switching",
                "targetRemoteSwitch": "remote_switching",
                "remoteSwitch": "remote_switching",
            }

            setattr(self.devices[device_uid].settings_property['protection'], switching_type[name], value)
            self._logger.debug("Updating %s protection of %s to %s", switching_type[name], element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value, switching_type[name]))

    def _remote_control(self, message: dict):
        """ Update a remote control. """
        element_uid = message['properties']['uid']
        key_pressed = message['properties']['property.value.new']

        # The message for the diary needs to be ignored
        if key_pressed is not None:
            device_uid = get_device_uid_from_element_uid(element_uid)
            old_key_pressed = self.devices[device_uid].remote_control_property[element_uid].key_pressed
            self.devices[device_uid].remote_control_property[element_uid].key_pressed = key_pressed
            self._logger.debug("Updating remote control of %s. Key %s",
                               element_uid,
                               f"pressed: {key_pressed}" if key_pressed != 0 else f"released: {old_key_pressed}")
            self._publisher.dispatch(device_uid, (element_uid, key_pressed))

    def _since_time(self, message: dict):
        """ Update point in time the total consumption was reset. """
        element_uid = message['uid']
        total_since = message['property.value.new']
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices[device_uid].consumption_property[element_uid].total_since = total_since
        self._logger.debug("Updating total since of %s to %s", element_uid, total_since)
        self._publisher.dispatch(device_uid, (element_uid, total_since, "total_since"))

    def _switch_type(self, message: dict):
        """ Update switch type setting (sts). """
        element_uid = message['properties']['uid']
        value = message['properties']['property.value.new'] * 2  # FWR, value.new is 1 for 2 buttons and 2 for 4 buttons.
        device_uid = get_device_uid_from_setting_uid(element_uid)
        self.devices[device_uid].settings_property['switch_type'].value = value
        self.devices[device_uid].remote_control_property[f'devolo.RemoteControl:{device_uid}'].key_count = value
        self._logger.debug("Updating switch type of %s to %s", device_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value))

    def _temperature_report(self, message: dict):
        """ Update temperature report settings. """
        if type(message['properties'].get("property.value.new")) not in [dict, list]:
            element_uid = message['properties']['uid']
            value = message['properties']['property.value.new']
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self.devices[device_uid].settings_property['temperature_report'].temp_report = value
            self._logger.debug("Updating temperature report of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _total_consumption(self, message: dict):
        """ Update total consumption. """
        self._update_consumption(element_uid=message['uid'], consumption="total", value=message['property.value.new'])

    def _unknown(self, message: dict):
        """ Ignore unknown messages. """
        ignore = (
            "devolo.DeviceEvents",
            "devolo.PairDevice",
            "devolo.SirenBinarySensor",
            "devolo.SirenMultiLevelSensor",
            "devolo.mprm.gw.GatewayManager",
            "devolo.mprm.gw.PortalManager",
            "ss",
            "mcs",
        )
        if not message["properties"]["uid"].startswith(ignore):
            self._logger.debug(json.dumps(message, indent=4))

    def _update_automatic_calibration(self, element_uid: str, calibration_status: bool):
        """ Update automatic calibration setting of a device. """
        device_uid = get_device_uid_from_setting_uid(element_uid)
        self.devices[device_uid].settings_property['automatic_calibration'].calibration_status = calibration_status
        self._logger.debug("Updating value of %s to %s", element_uid, calibration_status)
        self._publisher.dispatch(device_uid, (element_uid, calibration_status))

    def _update_consumption(self, element_uid: str, consumption: str, value: float):
        """ Update the consumption of a device. """
        device_uid = get_device_uid_from_element_uid(element_uid)
        setattr(self.devices[device_uid].consumption_property[element_uid], consumption, value)
        self._logger.debug("Updating %s consumption of %s to %s", consumption, element_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value, consumption))

    def _update_general_device_settings(self, element_uid: str, **kwargs: Any):
        """ Update general device settings. """
        device_uid = get_device_uid_from_setting_uid(element_uid)
        for key, value in kwargs.items():
            setattr(self.devices[device_uid].settings_property['general_device_settings'], key, value)
            self._logger.debug("Updating attribute: %s of %s to %s", key, element_uid, value)
            self._publisher.dispatch(device_uid, (key, value))

    def _voltage_multi_level_sensor(self, message: dict):
        """ Update a voltage value. """
        element_uid = message['properties']['uid']
        value = message['properties']['property.value.new']
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices[device_uid].multi_level_sensor_property[element_uid].value = value
        self._logger.debug("Updating voltage of %s to %s", element_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value))



import pytest


@pytest.fixture()
def mock_publisher_add_event(mocker):
    """ Mock add event function to keep tests independent from a real event. """
    mocker.patch("devolo_home_control_api.publisher.publisher.Publisher.add_event", return_value=None)


@pytest.fixture()
def mock_publisher_delete_event(mocker):
    """ Mock delete event function to keep tests independent from a real event. """
    mocker.patch("devolo_home_control_api.publisher.publisher.Publisher.delete_event", return_value=None)


@pytest.fixture()
def mock_publisher_dispatch(mocker):
    """ Mock dispatch function to keep tests independent from a real subscriber. """
    mocker.patch("devolo_home_control_api.publisher.publisher.Publisher.dispatch", return_value=None)



import time

import pytest
from requests.exceptions import ConnectionError
from urllib3.connection import ConnectTimeoutError
from websocket import WebSocketApp

from devolo_home_control_api.backend.mprm_rest import MprmRest
from devolo_home_control_api.backend.mprm_websocket import MprmWebsocket

from .mocks.mock_websocket import MockWebsocketError
from .stubs.mprm_websocket import StubMprmWebsocket


@pytest.mark.usefixtures("mprm_instance")
class TestMprmWebsocket:

    @pytest.mark.usefixtures("mock_mprmwebsocket_websocketapp")
    def test_websocket_connect(self, mprm_session, gateway_instance):
        self.mprm._session = mprm_session
        self.mprm.gateway = gateway_instance
        with pytest.raises(AssertionError):
            self.mprm.websocket_connect()

    def test_websocket_disconnect(self):
        self.mprm._ws = MockWebsocketError()
        with pytest.raises(AssertionError):
            self.mprm.websocket_disconnect()

    @pytest.mark.usefixtures("mock_mprmwebsocket_websocket_disconnect")
    def test__exit__(self, mocker):
        disconnect_spy = mocker.spy(MprmWebsocket, "websocket_disconnect")
        self.mprm.__exit__(None, None, None)
        assert disconnect_spy.call_count == 1

    @pytest.mark.usefixtures("mock_mprmwebsocket_on_update")
    def test__on_message_event_sequence(self):
        event_sequence = self.mprm._event_sequence
        self.mprm._event_sequence = 5
        message = '{"properties": {"com.prosyst.mbs.services.remote.event.sequence.number": 5}}'
        self.mprm._on_message(None, message)
        assert event_sequence != self.mprm._event_sequence
        assert self.mprm._event_sequence == 6
        message = '{"properties": {"com.prosyst.mbs.services.remote.event.sequence.number": 7}}'
        self.mprm._on_message(None, message)
        assert self.mprm._event_sequence == 8

    @pytest.mark.usefixtures("mock_mprmwebsocket_get_remote_session")
    @pytest.mark.usefixtures("mock_mprmwebsocket_websocket_connection")
    @pytest.mark.usefixtures("mock_mprmwebsocket_try_reconnect")
    @pytest.mark.usefixtures("mock_mprmwebsocket_websocketapp")
    def test__on_error(self, mocker):
        close_spy = mocker.spy(WebSocketApp, "close")
        reconnect_spy = mocker.spy(MprmWebsocket, "_try_reconnect")
        connect_spy = mocker.spy(MprmWebsocket, "websocket_connect")
        self.mprm._ws = WebSocketApp()
        self.mprm._on_error(self.mprm._ws, "error")
        assert close_spy.call_count == 1
        assert reconnect_spy.call_count == 1
        assert connect_spy.call_count == 1

    @pytest.mark.usefixtures("mock_mprmwebsocket_websocketapp")
    @pytest.mark.usefixtures("mock_session_get")
    def test__on_pong(self, mocker, mprm_session, gateway_instance, mydevolo):
        spy = mocker.spy(MprmRest, "refresh_session")
        self.mprm._mydevolo = mydevolo
        self.mprm._session = mprm_session
        self.mprm.gateway = gateway_instance
        self.mprm._local_ip = self.gateway["local_ip"]
        self.mprm._on_pong(None)
        assert spy.call_count == 1

    @pytest.mark.usefixtures("mock_mprmwebsocket_websocketapp")
    def test__try_reconnect(self, mocker):
        spy = mocker.spy(time, "sleep")
        self.mprm._ws = ConnectTimeoutError
        self.mprm._local_ip = self.gateway["local_ip"]
        self.mprm._try_reconnect(0.1)
        spy.assert_called_once_with(0.1)

    @pytest.mark.usefixtures("mock_mprmwebsocket_websocketapp")
    def test__try_reconnect_with_detect(self, mocker):
        spy_sleep = mocker.spy(time, "sleep")
        spy_detect_gateway = mocker.spy(StubMprmWebsocket, "detect_gateway_in_lan")
        self.mprm._ws = ConnectionError
        self.mprm._local_ip = self.gateway["local_ip"]
        self.mprm._try_reconnect(4)
        spy_sleep.assert_called_once_with(1)
        spy_detect_gateway.assert_called_once()



from datetime import datetime

import pytest
from devolo_home_control_api.backend import MESSAGE_TYPES


@pytest.mark.usefixtures("mock_inspect_devices_metering_plug")
@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_mydevolo__call")
class TestHomeControl:

    def test_hasattr(self):
        homecontrol_functions = (func for func in dir(self.homecontrol)
                                 if callable(getattr(self.homecontrol,
                                                     func)) and not func.startswith("__"))
        assert set(MESSAGE_TYPES.values()).difference(homecontrol_functions) == {"_gateway_accessible",
                                                                                 "_device_state"}

    def test_binary_sensor_devices(self):
        assert hasattr(self.homecontrol.binary_sensor_devices[0], "binary_sensor_property")

    def test_binary_switch_devices(self):
        assert hasattr(self.homecontrol.binary_switch_devices[0], "binary_switch_property")

    def test_blinds_devices(self):
        assert hasattr(self.homecontrol.blinds_devices[0], "multi_level_switch_property")

    def test_multi_level_sensor_devices(self):
        assert hasattr(self.homecontrol.multi_level_sensor_devices[0], "multi_level_sensor_property")

    def test_multi_level_switch_devices(self):
        assert hasattr(self.homecontrol.multi_level_switch_devices[0], "multi_level_switch_property")

    def test_remote_control_devices(self):
        assert hasattr(self.homecontrol.remote_control_devices[0], "remote_control_property")

    def test_get_publisher(self):
        assert len(self.homecontrol.publisher._events) == 10

    def test__automatic_calibration(self):
        device = self.devices['blinds']
        uid = device['uid']
        self.homecontrol._automatic_calibration({
            "UID": f"acs.{uid}",
            "properties": {
                "calibrationStatus": device['calibrationStatus']
            }
        })
        assert self.homecontrol.devices.get(uid).settings_property['automatic_calibration'].calibration_status == \
            bool(device['calibrationStatus'])

    def test___binary_sync(self):
        device = self.devices['blinds']
        uid = device['uid']
        self.homecontrol._binary_sync({
            "UID": f"acs.{uid}",
            "properties": {
                "value": device['inverted']
            }
        })
        assert self.homecontrol.devices.get(uid).settings_property['movement_direction'].inverted is device['inverted']

    def test__binary_async_blinds(self):
        device = self.devices.get("blinds").get("uid")
        i2 = self.devices.get("blinds").get("i2")
        self.homecontrol._binary_async({
            "UID": f"bas.{device}#i2",
            "properties": {
                "value": not i2
            }
        })
        assert self.homecontrol.devices.get(device).settings_property.get("i2").value is not i2

    def test__binary_async_siren(self):
        device = self.devices.get("siren").get("uid")
        muted = self.devices.get("siren").get("muted")
        self.homecontrol._binary_async({
            "UID": f"bas.{device}",
            "properties": {
                "value": not muted
            }
        })
        assert self.homecontrol.devices.get(device).settings_property.get("muted").value is not muted

    def test__binary_sensor(self):
        device = self.devices.get("sensor").get("uid")
        del self.homecontrol.devices[device].binary_sensor_property
        assert not hasattr(self.homecontrol.devices.get(device), "binary_sensor_property")
        self.homecontrol._binary_sensor({
            "UID": self.devices.get("sensor").get("elementUIDs")[0],
            "properties": {
                "state": self.devices.get("sensor").get("state"),
                "sensorType": self.devices.get("sensor").get("sensor_type"),
                "subType": ""
            },
        })
        assert hasattr(self.homecontrol.devices.get(device), "binary_sensor_property")

    def test__binary_switch(self):
        device = self.devices['mains']['uid']
        del self.homecontrol.devices[device].binary_switch_property
        assert not hasattr(self.homecontrol.devices[device], "binary_switch_property")
        self.homecontrol._binary_switch({
            "UID": self.devices['mains']['properties']['elementUIDs'][1],
            "properties": self.devices['mains']['properties'],
        })
        assert hasattr(self.homecontrol.devices[device], "binary_switch_property")

    def test__consumption(self):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[device].consumption_property
        assert not hasattr(self.homecontrol.devices.get(device), "consumption_property")
        self.homecontrol._meter({
            "UID": "devolo.Meter:hdm:ZWave:F6BF9812/2",
            "properties": {
                "currentValue": self.devices.get("mains").get("current_consumption"),
                "totalValue": self.devices.get("mains").get("total_consumption"),
                "sinceTime": self.devices.get("mains").get("properties").get("total_consumption"),
            }
        })
        assert hasattr(self.homecontrol.devices.get(device), "consumption_property")

    def test__general_device(self):
        device = self.devices['mains']
        element_uid = f"gds.{device['uid']}"
        self.homecontrol.devices[device['uid']].settings_property['general_device_settings'].events_enabled = \
            not device['properties']['eventsEnabled']
        self.homecontrol._general_device({
            "UID": element_uid,
            "properties": {
                "settings": {
                    "eventsEnabled": device['properties']['eventsEnabled'],
                    "name": device['properties']['itemName'],
                    "zoneID": device['properties']['zoneId'],
                    "icon": device['properties']['icon'],
                }
            }
        })
        assert self.homecontrol.devices[device['uid']].settings_property['general_device_settings'].events_enabled == \
            device['properties']['eventsEnabled']

    def test__humidity_bar(self):
        # TODO: Use test data
        device = self.devices.get("humidity").get("uid")
        del self.homecontrol.devices[device].humidity_bar_property
        assert not hasattr(self.homecontrol.devices.get(device), "humidity_bar_property")
        self.homecontrol._humidity_bar({
            "UID": f"devolo.HumidityBarValue:{device}",
            "properties": {
                "sensorType": "humidityBarPos",
                "value": 75,
            }
        })
        self.homecontrol._humidity_bar({
            "UID": f"devolo.HumidityBarZone:{device}",
            "properties": {
                "sensorType": "humidityBarZone",
                "value": 1,
            }
        })
        assert self.homecontrol.devices.get(device).humidity_bar_property.get(f"devolo.HumidityBar:{device}").value == 75
        assert self.homecontrol.devices.get(device).humidity_bar_property.get(f"devolo.HumidityBar:{device}").zone == 1

    def test__last_activity_binary_sensor(self):
        device = self.devices['sensor']['uid']
        element_uids = self.devices['sensor']['elementUIDs']
        self.homecontrol._binary_sensor({
            "UID": element_uids[0],
            "properties": {
                "state": self.devices['sensor']['state'],
                "sensorType": self.devices['sensor']['sensor_type'],
                "subType": "",
            }
        })
        self.homecontrol._last_activity({
            "UID": element_uids[1],
            "properties": {
                "lastActivityTime": self.devices['sensor']['last_activity'],
            }
        })
        assert self.homecontrol.devices[device].binary_sensor_property.get(element_uids[0]).last_activity == \
            datetime.utcfromtimestamp(self.devices['sensor']['last_activity'] / 1000)

    def test__last_activity_siren(self):
        device = self.devices['siren']['uid']
        element_uids = self.devices['siren']['elementUIDs']
        self.homecontrol._multi_level_switch({
            "UID": element_uids[0],
            "properties": {
                "value": self.devices['siren']['properties']['value'],
                "switchType": self.devices['siren']['switch_type'],
                "max": self.devices['siren']['properties']['max'],
                "min": self.devices['siren']['properties']['min'],
            }
        })
        self.homecontrol._last_activity({
            "UID": element_uids[3],
            "properties": {
                "lastActivityTime": self.devices['siren']['last_activity'],
            }
        })
        assert self.homecontrol.devices[device].multi_level_switch_property[element_uids[0]].last_activity == \
            datetime.utcfromtimestamp(self.devices['siren']['last_activity'] / 1000)

    def test__led(self):
        device = self.devices['mains']
        element_uid = f"lis.{device['UID']}"
        led_setting = device['properties']['led_setting']
        self.homecontrol._led({
            "UID": element_uid,
            "properties": {
                "led": led_setting,
            }
        })
        assert self.homecontrol.devices[device['UID']].settings_property['led'].led_setting == led_setting
        device = self.devices['sensor']
        element_uid = f"vfs.{device['UID']}"
        led_setting = device['led_setting']
        self.homecontrol._led({
            "UID": element_uid,
            "properties": {
                "feedback": led_setting,
            }
        })
        assert self.homecontrol.devices[device['UID']].settings_property['led'].led_setting == led_setting

    def test__multilevel_async(self):
        device = self.devices['blinds']
        uid = device['uid']
        self.homecontrol._multilevel_async({
            "UID": f"mas.{uid}",
            "properties": {
                "itemId": "motorActivity",
                "value": device['motorActivity'],
            }
        })
        assert self.homecontrol.devices[uid].settings_property['motor_activity'].value == device['motorActivity']

    def test__multilevel_async_mains(self):
        device = self.devices['mains']
        uid = device['uid']
        self.homecontrol._multilevel_async({
            "UID": f"mas.{uid}",
            "properties": {
                "itemId": None,
                "value": device['flashMode'],
            }
        })
        assert self.homecontrol.devices[uid].settings_property['flash_mode'].value == device['flashMode']

    def test__multilevel_async_type_error(self):
        with pytest.raises(TypeError):
            device = self.devices['sensor']
            uid = device['uid']
            self.homecontrol._multilevel_async({
                "UID": f"mas.{uid}",
                "properties": {
                    "itemId": None,
                    "value": device['motion_sensitivity'],
                }
            })

    def test__multilevel_sync_sensor(self):
        device = self.devices['sensor']['uid']
        self.homecontrol._multilevel_sync({
            "UID": self.devices['sensor']['settingUIDs'][2],
            "properties": self.devices['sensor']['properties'],
        })
        assert self.homecontrol.devices[device].settings_property['motion_sensitivity'].motion_sensitivity == \
            self.devices['sensor']['properties']['value']

    def test__multilevel_sync_shutter(self):
        device = self.devices['blinds']['uid']
        self.homecontrol._multilevel_sync({
            "UID": f"mss.{device}",
            "properties": {
                "value": self.devices['blinds']['shutter_duration'],
            }
        })
        assert self.homecontrol.devices[device].settings_property['shutter_duration'].shutter_duration == \
            self.devices['blinds']['shutter_duration']

    def test__multilevel_sync_siren(self):
        device = self.devices['siren']['uid']
        self.homecontrol._multilevel_sync({
            "UID": self.devices['siren']['settingUIDs'][0],
            "properties": self.devices['siren']['properties'],
        })
        assert self.homecontrol.devices[device].settings_property['tone'].tone == self.devices['siren']['properties']['value']

    def test__multi_level_sensor(self):
        # TODO: Use test data
        device = self.devices.get("sensor").get("uid")
        del self.homecontrol.devices[device].multi_level_sensor_property
        assert not hasattr(self.homecontrol.devices.get(device), "multi_level_sensor_property")
        self.homecontrol._multi_level_sensor({
            "UID": self.devices.get("sensor").get("elementUIDs")[2],
            "properties": {
                "value": 90.0,
                "unit": "%",
                "sensorType": "light",
            }
        })
        assert hasattr(self.homecontrol.devices.get(device), "multi_level_sensor_property")

    def test__multi_level_switch(self):
        device = self.devices.get("siren").get("uid")
        del self.homecontrol.devices[device].multi_level_switch_property
        assert not hasattr(self.homecontrol.devices.get(device), "multi_level_switch_property")
        self.homecontrol._multi_level_switch({
            "UID": self.devices.get("siren").get("elementUIDs")[0],
            "properties": {
                "state": self.devices.get("multi_level_switch").get("state"),
                "value": self.devices.get("multi_level_switch").get("value"),
                "switchType": self.devices.get("multi_level_switch").get("switch_type"),
                "max": self.devices.get("multi_level_switch").get("max"),
                "min": self.devices.get("multi_level_switch").get("min"),
            }
        })
        assert hasattr(self.homecontrol.devices.get(device), "multi_level_switch_property")

    def test__parameter(self):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._parameter({
            "UID": "cps.hdm:ZWave:F6BF9812/2",
            "properties": {
                "paramChanged": False,
            }
        })
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("param_changed"), "param_changed")

    def test__protection(self):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._protection({
            "UID": "ps.hdm:ZWave:F6BF9812/2",
            "properties": {
                "localSwitch": True,
                "remoteSwitch": False,
            }
        })
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("protection"), "local_switching")
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("protection"), "remote_switching")

    def test__remote_control(self):
        device = self.devices.get("remote").get("uid")
        element_uid = self.devices.get("remote").get("elementUIDs")[0]
        del self.homecontrol.devices[device].remote_control_property
        assert not hasattr(self.homecontrol.devices.get(device), "remote_control_property")
        self.homecontrol._remote_control({
            "UID": element_uid,
            "properties": {
                "keyCount": self.devices.get("remote").get("key_count"),
                "keyPressed": 0,
                "type": 1,
            }
        })
        assert self.homecontrol.devices.get(device).remote_control_property[element_uid].key_pressed == 0

    def test__switch_type(self):
        device = self.devices['remote']['uid']
        self.homecontrol._switch_type({
            "UID": f"sts.{device}",
            "properties": {
                "switchType": self.devices['remote']['key_count'] / 2,
            }
        })
        assert self.homecontrol.devices[device].settings_property['switch_type'].value == self.devices['remote']['key_count']

    def test__temperature_report(self):
        # TODO: Use test data
        device = self.devices.get("sensor").get("uid")
        self.homecontrol._temperature_report({
            "UID": "trs.hdm:ZWave:F6BF9812/6",
            "properties": {
                "tempReport": True,
                "targetTempReport": False,
            }
        })
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("temperature_report"), "temp_report")
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("temperature_report"), "target_temp_report")

    @pytest.mark.usefixtures("mock_inspect_devices")
    def test_device_change_add(self, mocker):
        uids = [self.devices.get(device).get("uid") for device in self.devices]
        uids.append("test_uid")
        spy = mocker.spy(self.homecontrol, "_inspect_devices")
        self.homecontrol.device_change(uids)
        spy.assert_called_once_with(['test_uid'])

    def test_device_change_remove(self):
        uids = [self.devices.get(device).get("uid") for device in self.devices]
        del uids[4]
        self.homecontrol.device_change(uids)
        assert self.devices.get("mains").get("uid") not in self.homecontrol.devices.keys()

    @pytest.mark.usefixtures("mock_extract_data_from_element_uids")
    @pytest.mark.usefixtures("mock_mprmrest_get_all_devices")
    def test__inspect_devices(self):
        uid = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[uid]
        self.homecontrol._inspect_devices(self.homecontrol.get_all_devices())
        try:
            self.homecontrol.devices[uid]
            assert True
        except KeyError:
            assert False



MESSAGE_TYPES = {
    "devolo.BinarySensor": "_binary_sensor",
    "devolo.BinarySwitch": "_binary_switch",
    "devolo.Blinds": "_multi_level_switch",
    "devolo.DevicesPage": "_inspect_devices",
    "devolo.DewpointSensor": "_multi_level_sensor",
    "devolo.Dimmer": "_multi_level_switch",
    "devolo.Grouping": "_grouping",
    "devolo.HumidityBarValue": "_humidity_bar",
    "devolo.HumidityBarZone": "_humidity_bar",
    "devolo.mprm.gw.GatewayAccessibilityFI": "_gateway_accessible",
    "devolo.Meter": "_meter",
    "devolo.MildewSensor": "_binary_sensor",
    "devolo.MultiLevelSensor": "_multi_level_sensor",
    "devolo.MultiLevelSwitch": "_multi_level_switch",
    "devolo.RemoteControl": "_remote_control",
    "devolo.SirenMultiLevelSwitch": "_multi_level_switch",
    "devolo.ShutterMovementFI": "_binary_sensor",
    "devolo.ValveTemperatureSensor": "_multi_level_sensor",
    "devolo.VoltageMultiLevelSensor": "_multi_level_sensor",
    "devolo.WarningBinaryFI": "_binary_sensor",
    "hdm": "_device_state",
    "acs.hdm": "_automatic_calibration",
    "bas.hdm": "_binary_async",
    "bss.hdm": "_binary_sync",
    "cps.hdm": "_parameter",
    "gds.hdm": "_general_device",
    "lis.hdm": "_led",
    "mas.hdm": "_multilevel_async",
    "mss.hdm": "_multilevel_sync",
    "ps.hdm": "_protection",
    "stmss.hdm": "_multilevel_sync",
    "sts.hdm": "_switch_type",
    "trs.hdm": "_temperature_report",
    "vfs.hdm": "_led"
}



import json
import pathlib

from devolo_home_control_api.mydevolo import Mydevolo


class MockGateway:

    def __init__(self, gateway_id: str, mydevolo: Mydevolo):
        file = pathlib.Path(__file__).parent / ".." / "test_data.json"
        with file.open("r") as fh:
            test_data = json.load(fh)

        self.id = gateway_id
        self.name = test_data.get("gateway").get("name")
        self.role = test_data.get("gateway").get("role")
        self.local_user = test_data.get("user").get("uuid")
        self.local_passkey = test_data.get("gateway").get("local_passkey")
        self.full_url = test_data.get("gateway").get("full_url")
        self.external_access = test_data.get("gateway").get("external_access")
        self.status = test_data.get("gateway").get("status")
        self.state = test_data.get("gateway").get("state")
        self.firmware_version = test_data.get("gateway").get("firmware_version")
        self.online = True
        self.zones = test_data.get("gateway").get("zones")

    def update_state(self, online: bool):
        self.online = online



from json import JSONDecodeError

from requests.exceptions import ConnectTimeout, ReadTimeout


class MockResponse:

    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code
        self.ok = not 400 <= self.status_code <= 600
        self.url = "https://homecontrol.mydevolo.com"


class MockResponseConnectTimeout(MockResponse):

    def get(self, url, auth=None, timeout=None):
        raise ConnectTimeout


class MockResponseGet(MockResponse):

    def get(self, url, auth=None, timeout=None):
        return MockResponseGet({"link": "test_link"},
                               status_code=200)

    def json(self):
        return self.json_data


class MockResponseJsonError(MockResponse):

    def get(self, url, auth=None, timeout=None):
        raise JSONDecodeError(msg="message", doc="doc", pos=1)


class MockResponsePost(MockResponse):

    def post(self, url, auth=None, timeout=None, headers=None, data=None):
        return MockResponsePost({"link": "test_link"},
                                status_code=200)

    def json(self):
        return {
            "id": 2
        }


class MockResponseReadTimeout(MockResponse):

    def post(self, url, auth=None, timeout=None, headers=None, data=None):
        raise ReadTimeout


class MockResponseServiceUnavailable(MockResponse):

    def get(self, url, auth=None, timeout=None):
        return MockResponseGet({"link": "test_link"},
                               status_code=503)

    def json(self):
        return self.json_data



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.multi_level_switch_property import MultiLevelSwitchProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def shutter(device_uid: str) -> Zwave:
    """
    Represent a shutter in tests

    :param device_uid: Device UID this mock shall have
    :return: Metering Plug device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data['devices']['blinds'])

    device.multi_level_switch_property = {}
    device.settings_property = {}

    device.multi_level_switch_property[f'devolo.Blinds:{device_uid}'] = MultiLevelSwitchProperty(
        element_uid=f"devolo.Blinds:{device_uid}",
        setter=lambda uid,
        state: None,
        value=test_data['devices']['blinds']['value'],
        max=test_data['devices']['blinds']['max'],
        min=test_data['devices']['blinds']['min'])

    device.settings_property['i2'] = SettingsProperty(element_uid=f"bas.{device_uid}",
                                                      setter=lambda uid,
                                                      state: None,
                                                      value=test_data['devices']['blinds']['i2'])

    device.settings_property["general_device_settings"] = SettingsProperty(element_uid=f'gds.{device_uid}',
                                                                           setter=lambda uid,
                                                                           state: None,
                                                                           icon=test_data['devices']['blinds']['icon'],
                                                                           name=test_data['devices']['blinds']['itemName'],
                                                                           zone_id=test_data['devices']['blinds']['zoneId'],
                                                                           zones=test_data['gateway']['zones'])

    device.settings_property["automatic_calibration"] = SettingsProperty(
        element_uid=f'acs.{device_uid}',
        setter=lambda uid,
        state: None,
        calibration_status=test_data['devices']['blinds']['calibrationStatus'] == 2)

    device.settings_property["movement_direction"] = SettingsProperty(
        element_uid=f'bss.{device_uid}',
        setter=lambda uid,
        state: None,
        direction=not bool(test_data['devices']['blinds']['movement_direction']))

    device.settings_property["shutter_duration"] = SettingsProperty(
        element_uid=f'mss.{device_uid}',
        setter=lambda uid,
        state: None,
        shutter_duration=test_data['devices']['blinds']['shutter_duration'])

    return device



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty
from devolo_home_control_api.properties.consumption_property import ConsumptionProperty
from devolo_home_control_api.properties.multi_level_sensor_property import MultiLevelSensorProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def metering_plug(device_uid: str) -> Zwave:
    """
    Represent a metering plug in tests

    :param device_uid: Device UID this mock shall have
    :return: Metering Plug device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data['devices']['mains']['properties'])

    device.binary_switch_property = {}
    device.consumption_property = {}
    device.multi_level_sensor_property = {}
    device.settings_property = {}

    device.binary_switch_property[f'devolo.BinarySwitch:{device_uid}'] = BinarySwitchProperty(
        element_uid=f"devolo.BinarySwitch:{device_uid}",
        setter=lambda uid,
        state: None,
        state=test_data['devices']['mains']['properties']['state'],
        enabled=test_data['devices']['mains']['properties']['guiEnabled'])
    device.consumption_property[f'devolo.Meter:{device_uid}'] = ConsumptionProperty(
        element_uid=f"devolo.Meter:{device_uid}",
        setter=lambda uid,
        state: None,
        current=test_data['devices']['mains']['properties']['current_consumption'],
        total=test_data['devices']['mains']['properties']['total_consumption'],
        total_since=test_data['devices']['mains']['properties']['sinceTime'])
    device.multi_level_sensor_property[f'devolo.VoltageMultiLevelSensor:{device_uid}'] = \
        MultiLevelSensorProperty(element_uid=f"devolo.VoltageMultiLevelSensor:{device_uid}",
                                 current=test_data['devices']['mains']['properties']['voltage'])
    device.settings_property["param_changed"] = SettingsProperty(element_uid=f"cps.{device_uid}",
                                                                 setter=lambda uid,
                                                                 state: None)
    device.settings_property['general_device_settings'] = SettingsProperty(
        element_uid=f"gds.{device_uid}",
        setter=lambda uid,
        state: True,
        events_enabled=test_data['devices']['mains']['properties']['eventsEnabled'],
        icon=test_data['devices']['mains']['properties']['icon'],
        name=test_data['devices']['mains']['properties']['itemName'],
        zone_id=test_data['devices']['mains']['properties']['zoneId'],
        zones=test_data.get("gateway").get("zones"))
    device.settings_property["led"] = SettingsProperty(element_uid=f"lis.{device_uid}",
                                                       setter=lambda uid,
                                                       state: True,
                                                       led_setting=test_data['devices']['mains']['properties']['led_setting'])
    device.settings_property["protection"] = \
        SettingsProperty(element_uid=f"ps.{device_uid}",
                         setter=lambda uid, state: True,
                         local_switching=test_data['devices']['mains']['properties']['local_switch'],
                         remote_switching=test_data['devices']['mains']['properties']['remote_switch'])
    device.settings_property['flash_mode'] = SettingsProperty(element_uid=f"mas.{device_uid}",
                                                              setter=lambda uid,
                                                              state: None,
                                                              valus=test_data['devices']['mains']['flashMode'])

    return device



import json
import pathlib

from requests import Session

from devolo_home_control_api.backend.mprm import Mprm
from devolo_home_control_api.mydevolo import Mydevolo

from ..mocks.mock_gateway import MockGateway


class StubMprm(Mprm):

    def __init__(self):
        file = pathlib.Path(__file__).parent / ".." / "test_data.json"
        with file.open("r") as fh:
            test_data = json.load(fh)

        self._mydevolo = Mydevolo()
        self._session = Session()
        self._zeroconf = None
        self.gateway = MockGateway(test_data['gateway']['id'], self._mydevolo)
        super().__init__()

    def on_update(self, message):
        pass



from abc import ABC

from .property import Property


class SensorProperty(Property, ABC):
    """
    Abstract object for sensors. It stores the sensor and sub type.

    :param connection: Collection of instances needed to communicate with the central unit
    :param element_uid: Element UID
    :key sensor_type: Type of the sensor sensor, something like 'alarm'
    :type sensor_type: str
    :key sub_type: Subtype of the sensor, something like 'overload'
    :type sub_type: str
    """

    def __init__(self, element_uid: str, **kwargs: str):
        super().__init__(element_uid=element_uid)

        self.sensor_type: str = kwargs.pop("sensor_type", "")
        self.sub_type: str = kwargs.pop("sub_type", "")



from datetime import datetime

from ..exceptions.device import WrongElementError
from .sensor_property import SensorProperty


class HumidityBarProperty(SensorProperty):
    """
    Object for humidity bars. It stores the zone and the position inside that zone.

    :param element_uid: Fake element UID, something like devolo.HumidityBar:hdm:ZWave:CBC56091/24
    :key value: Position inside a zone
    :type value: int
    :key zone: Humidity zone from 0 (very dry) to 4 (very humid)
    :type zone: int
    """

    def __init__(self, element_uid: str, **kwargs):
        if not element_uid.startswith("devolo.HumidityBar:"):
            raise WrongElementError(f"{element_uid} is not a humidity bar.")

        super().__init__(element_uid=element_uid, **kwargs)

        self._value: int = kwargs.pop("value", 0)
        self.zone: int = kwargs.pop("zone", 0)

    @property
    def value(self) -> int:
        """ Position inside a zone. """
        return self._value

    @value.setter
    def value(self, value: int):
        """
        Update value and set point in time of the last_activity. As zone never changes without a value, just the value sets
        the last activity.
        """
        self._value = value
        self._last_activity = datetime.now()
        self._logger.debug("value of element_uid %s set to %s.", self.element_uid, value)



import pytest


@pytest.mark.usefixtures("mock_inspect_devices_metering_plug")
@pytest.mark.usefixtures("home_control_instance")
class TestPublisher:

    def test_add_event(self):
        self.homecontrol.publisher.add_event("Test")
        assert "Test" in self.homecontrol.publisher._events

    def test_delete_event(self):
        self.homecontrol.publisher._events = {
            "Test": {}
        }
        self.homecontrol.publisher.delete_event("Test")
        assert "Test" not in self.homecontrol.publisher._events

    def test_register_unregister(self):
        for device in self.homecontrol.devices:
            self.homecontrol.devices[device].subscriber = Subscriber(device)
            self.homecontrol.publisher.register(device, self.homecontrol.devices[device].subscriber)
            assert len(self.homecontrol.publisher._get_subscribers_for_specific_event(device)) == 1
        for device in self.homecontrol.devices:
            self.homecontrol.publisher.unregister(device, self.homecontrol.devices[device].subscriber)
            assert len(self.homecontrol.publisher._get_subscribers_for_specific_event(device)) == 0

    def test_dispatch(self):
        for device in self.homecontrol.devices:
            self.homecontrol.devices[device].subscriber = Subscriber(device)
            self.homecontrol.publisher.register(device, self.homecontrol.devices[device].subscriber)
        with pytest.raises(FileExistsError):
            self.homecontrol.publisher.dispatch(event="hdm:ZWave:F6BF9812/4", message=())


class Subscriber:

    def __init__(self, name):
        self.name = name

    def update(self, message):
        # We raise an error here so we can check for it in the test case.
        raise FileExistsError



import logging

from devolo_home_control_api.homecontrol import HomeControl
from devolo_home_control_api.mydevolo import Mydevolo

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(name)s %(levelname)s: %(message)s")

user = "username"
password = "password"


class Subscriber:

    def __init__(self, name):
        self.name = name

    def update(self, message):
        print(f'{self.name} got message "{message}"')


mydevolo = Mydevolo()
mydevolo.user = user
mydevolo.password = password

gateway_id = mydevolo.get_gateway_ids()[0]
homecontrol = HomeControl(gateway_id=gateway_id, mydevolo_instance=mydevolo)

for device in homecontrol.devices:
    homecontrol.devices[device].subscriber = Subscriber(device)
    homecontrol.publisher.register(device, homecontrol.devices[device].subscriber)
